import decimal
from typing import TypeAlias

# TODO: replace with pep 695 type alias once cattrs 24.1 is released
TimeStep: TypeAlias = decimal.Decimal
"""A type alias that represents the type of a time step.

A time step is represented as a decimal number to avoid floating point errors.
"""



from numbers import Real
from typing import overload, Optional

import numpy as np

from .units import Unit, Quantity, dimensionless


def base_units(units: Unit) -> Unit:
    """Return the base units of the given units.

    Args:
        units: The units to convert to base units.

    Returns:
        The base units of the given units.
    """

    return (1 * units).to_base_units().units


def is_in_base_units(units: Unit) -> bool:
    """Check if the units is only expressed in terms of base SI units.

    For example, `kg`, `m/s` are expressed in terms of base units, but `mg`, `km/h` or
    `dB` are not.

    Args:
        units: The units to check.

    Returns:
        True if the units are expressed in the base units of the registry.
    """

    return base_units(units) == units


@overload
def convert_to_base_units[
    T: (Real, np.ndarray)
](magnitude: T, units: None) -> tuple[T, None]: ...


@overload
def convert_to_base_units[
    T: (Real, np.ndarray)
](magnitude: T, units: Unit) -> tuple[np.ndarray, Optional[Unit]]: ...


def convert_to_base_units(magnitude, units):
    if units is None:
        return magnitude, None
    else:
        quantity = Quantity(magnitude, units)
        in_base_units = quantity.to_base_units()
        magnitude_in_base_units = in_base_units.magnitude
        base_units = in_base_units.units
        if base_units == dimensionless:
            base_units = None
        else:
            assert is_in_base_units(base_units)
        return magnitude_in_base_units, base_units



from collections.abc import Mapping

import numpy
import polars
import scipy.optimize

from .stats import is_error_dtype, VALUE_FIELD, ERROR_FIELD


def curve_fit(f, x_data, y_data, initial_values: Mapping[str, float]):
    x = numpy.array(x_data, dtype=numpy.float64)
    if is_error_dtype(y_data.dtype):
        y = numpy.array(y_data.struct.field("value"), dtype=numpy.float64)
        sigma = numpy.array(y_data.struct.field("error"), dtype=numpy.float64)
        if numpy.any(sigma == 0):
            raise ValueError("One of the error on y_data is 0")
    else:
        y = numpy.array(y_data, dtype=numpy.float64)
        sigma = None
    popt, pcov = scipy.optimize.curve_fit(
        f,
        x,
        y,
        p0=list(initial_values.values()),
        sigma=sigma,
        absolute_sigma=True,
        check_finite=True,
    )
    errors = numpy.sqrt(numpy.diag(pcov))
    return polars.DataFrame(
        [
            polars.Series(name, [{VALUE_FIELD: value, ERROR_FIELD: error}])
            for name, value, error in zip(initial_values, popt, errors)
        ]
    )



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'device_configurations_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class Ui_DeviceConfigurationsDialog(object):
    def setupUi(self, DeviceConfigurationsDialog):
        if not DeviceConfigurationsDialog.objectName():
            DeviceConfigurationsDialog.setObjectName("DeviceConfigurationsDialog")
        DeviceConfigurationsDialog.resize(620, 319)
        self.verticalLayout = QVBoxLayout(DeviceConfigurationsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.add_device_button = QToolButton(DeviceConfigurationsDialog)
        self.add_device_button.setObjectName("add_device_button")

        self.horizontalLayout.addWidget(self.add_device_button)

        self.remove_device_button = QToolButton(DeviceConfigurationsDialog)
        self.remove_device_button.setObjectName("remove_device_button")

        self.horizontalLayout.addWidget(self.remove_device_button)

        self.buttonBox = QDialogButtonBox(DeviceConfigurationsDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(
            QDialogButtonBox.Cancel | QDialogButtonBox.Save
        )

        self.horizontalLayout.addWidget(self.buttonBox)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(DeviceConfigurationsDialog)
        self.buttonBox.accepted.connect(DeviceConfigurationsDialog.accept)
        self.buttonBox.rejected.connect(DeviceConfigurationsDialog.reject)

        QMetaObject.connectSlotsByName(DeviceConfigurationsDialog)

    # setupUi

    def retranslateUi(self, DeviceConfigurationsDialog):
        DeviceConfigurationsDialog.setWindowTitle(
            QCoreApplication.translate(
                "DeviceConfigurationsDialog", "Edit device configurations...", None
            )
        )
        self.add_device_button.setText(
            QCoreApplication.translate("DeviceConfigurationsDialog", "...", None)
        )
        self.remove_device_button.setText(
            QCoreApplication.translate("DeviceConfigurationsDialog", "...", None)
        )

    # retranslateUi



from __future__ import annotations

import functools
from collections.abc import Callable

import anyio
import anyio.to_thread
from PySide6.QtCore import QSettings, Qt, QTimer
from PySide6.QtWidgets import (
    QMainWindow,
    QDockWidget,
    QDialog,
)

from caqtus.experiment_control.manager import ExperimentManager, Procedure
from caqtus.gui.common.exception_tree import ExceptionDialog
from caqtus.gui.common.waiting_widget import run_with_wip_widget
from caqtus.gui.condetrol._parameter_tables_editor import ParameterNamespaceEditor
from caqtus.session import ExperimentSessionMaker, PureSequencePath
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.recoverable_exceptions import (
    split_recoverable,
    SequenceInterruptedException,
)
from ._main_window_ui import Ui_CondetrolMainWindow
from .._extension import CondetrolExtensionProtocol
from .._logger import logger
from .._path_view import EditablePathHierarchyView
from .._sequence_widget import SequenceWidget
from ..device_configuration_editors._configurations_editor import (
    DeviceConfigurationsDialog,
)


class CondetrolWindowHandler:
    def __init__(
        self, main_window: CondetrolMainWindow, session_maker: ExperimentSessionMaker
    ):
        self.main_window = main_window
        self.session_maker = session_maker
        self.task_group = anyio.create_task_group()
        self.is_running_sequence = False

        self.main_window.path_view.sequence_start_requested.connect(self.start_sequence)
        self.main_window.sequence_widget.sequence_start_requested.connect(
            self.start_sequence
        )

    async def run_async(self) -> None:
        """Run the main window asynchronously."""

        async with self.task_group:
            self.task_group.start_soon(self.main_window.path_view.run_async)
            self.task_group.start_soon(self._monitor_global_parameters)
            self.task_group.start_soon(self.main_window.sequence_widget.exec_async)

    async def _monitor_global_parameters(self) -> None:
        while True:
            async with self.session_maker.async_session() as session:
                parameters = await session.get_global_parameters()
            if parameters != self.main_window.global_parameters_editor.get_parameters():
                self.main_window.global_parameters_editor.set_parameters(parameters)
                self.main_window.sequence_widget.set_available_parameter_names(
                    parameters.names()
                )
            await anyio.sleep(0.2)

    def start_sequence(self, path: PureSequencePath):
        try:
            experiment_manager = run_with_wip_widget(
                self.main_window,
                "Connecting to experiment manager...",
                self.main_window.connect_to_experiment_manager,
            )
        except Exception as e:
            logger.error("Failed to connect to experiment manager.", exc_info=e)
            self.main_window.display_error(
                "Failed to connect to experiment manager.", e
            )
            return

        if self.is_running_sequence:
            self.main_window.display_error(
                "A sequence is already running.",
                RuntimeError("A sequence is already running."),
            )
            return
        procedure = experiment_manager.create_procedure(
            "sequence launched from GUI", acquisition_timeout=1
        )
        self.task_group.start_soon(self._run_sequence, procedure, path)
        self.is_running_sequence = True

    async def _run_sequence(self, procedure: Procedure, sequence):
        with procedure:
            try:
                await anyio.to_thread.run_sync(procedure.start_sequence, sequence)
            except Exception as e:
                exception = RuntimeError(
                    f"An error occurred while starting the sequence {sequence}."
                )
                exception.__cause__ = e
                self.main_window.signal_exception_while_running_sequence(exception)
                return

            while await anyio.to_thread.run_sync(procedure.is_running_sequence):
                await anyio.sleep(50e-3)

            if (exc := procedure.exception()) is not None:
                # Here we ignore the SequenceInterruptedException because it is
                # expected to happen when the sequence is interrupted and we don't
                # want to display it to the user as an actual error.
                if isinstance(exc, SequenceInterruptedException):
                    exc = None
                elif isinstance(exc, ExceptionGroup):
                    _, exc = exc.split(SequenceInterruptedException)
                if exc is not None:
                    self.main_window.signal_exception_while_running_sequence(exc)
        self.is_running_sequence = False


class CondetrolMainWindow(QMainWindow, Ui_CondetrolMainWindow):
    """The main window of the Condetrol GUI.

    Parameters
    ----------
    session_maker
        A callable that returns an ExperimentSession.
        This is used to access the storage in which to look for sequences to display
        and edit.
    connect_to_experiment_manager
        A callable that is called to connect to an experiment manager in charge of
        running sequences.
        This is used to submit sequences to the manager when the user starts them
        in the GUI.
    extension
        The extension that provides the GUI with the necessary tools to edit sequences
        and device configurations.
    """

    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        connect_to_experiment_manager: Callable[[], ExperimentManager],
        extension: CondetrolExtensionProtocol,
    ):
        super().__init__()
        self.path_view = EditablePathHierarchyView(session_maker, self)
        self.global_parameters_editor = ParameterNamespaceEditor()
        self.connect_to_experiment_manager = connect_to_experiment_manager
        self.session_maker = session_maker
        self.sequence_widget = SequenceWidget(
            self.session_maker, extension.lane_extension, parent=self
        )
        self.device_configurations_dialog = DeviceConfigurationsDialog(
            extension.device_extension, parent=self
        )
        self.setup_ui()
        self.restore_window()
        self.setup_connections()
        self.timer = QTimer(self)

    def setup_ui(self):
        self.setupUi(self)
        self.setCentralWidget(self.sequence_widget)
        paths_dock = QDockWidget("Sequences", self)
        paths_dock.setObjectName("SequencesDock")
        paths_dock.setWidget(self.path_view)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, paths_dock)
        self.dock_menu.addAction(paths_dock.toggleViewAction())
        global_parameters_dock = QDockWidget("Global parameters", self)
        global_parameters_dock.setWidget(self.global_parameters_editor)
        global_parameters_dock.setObjectName("GlobalParametersDock")
        self.addDockWidget(
            Qt.DockWidgetArea.RightDockWidgetArea, global_parameters_dock
        )
        self.dock_menu.addAction(global_parameters_dock.toggleViewAction())
        # We hide the global parameters dock by default to reduce clutter when
        # launching the app the first time.
        global_parameters_dock.hide()

    def setup_connections(self):
        self.action_edit_device_configurations.triggered.connect(
            self.open_device_configurations_editor
        )
        self.path_view.sequence_double_clicked.connect(self.set_edited_sequence)
        self.path_view.sequence_interrupt_requested.connect(self.interrupt_sequence)
        self.global_parameters_editor.parameters_edited.connect(
            self._on_global_parameters_edited
        )

    def set_edited_sequence(self, path: PureSequencePath):
        self.sequence_widget.set_sequence(path)

    def on_procedure_exception(self, exception: Exception):
        recoverable, non_recoverable = split_recoverable(exception)
        if recoverable:
            logger.warning(
                f"Recoverable exception occurred while running a sequence",
                exc_info=recoverable,
            )
        if non_recoverable:
            # The exception will be logged anyway when condetrol crashes, so we don't
            # need to log it here.
            raise non_recoverable

        assert recoverable is not None

        self.display_error(
            f"An error occurred while running a sequence.",
            recoverable,
        )

    def open_device_configurations_editor(self) -> None:
        with self.session_maker() as session:
            previous_device_configurations = dict(session.default_device_configurations)
        self.device_configurations_dialog.set_device_configurations(
            previous_device_configurations
        )
        if self.device_configurations_dialog.exec() == QDialog.DialogCode.Accepted:
            new_device_configurations = (
                self.device_configurations_dialog.get_device_configurations()
            )
            with self.session_maker() as session:
                for device_name in session.default_device_configurations:
                    if device_name not in new_device_configurations:
                        del session.default_device_configurations[device_name]
                for (
                    device_name,
                    device_configuration,
                ) in new_device_configurations.items():
                    session.default_device_configurations[device_name] = (
                        device_configuration
                    )

    def closeEvent(self, a0):
        self.save_window()
        super().closeEvent(a0)

    def restore_window(self) -> None:
        ui_settings = QSettings()
        state = ui_settings.value(f"{__name__}/state", defaultValue=None)
        if state is not None:
            self.restoreState(state)
        geometry = ui_settings.value(f"{__name__}/geometry", defaultValue=None)
        if geometry is not None:
            self.restoreGeometry(geometry)

    def save_window(self) -> None:
        ui_settings = QSettings()
        ui_settings.setValue(f"{__name__}/state", self.saveState())
        ui_settings.setValue(f"{__name__}/geometry", self.saveGeometry())

    def display_error(self, message: str, exception: BaseException):
        exception_dialog = ExceptionDialog(self)
        exception_dialog.set_exception(exception)
        exception_dialog.set_message(message)
        exception_dialog.exec()

    def interrupt_sequence(self, path: PureSequencePath):
        experiment_manager = run_with_wip_widget(
            self,
            "Connecting to experiment manager...",
            self.connect_to_experiment_manager,
        )
        # we're actually lying here because we interrupt the running procedure, which
        # may be different from the one passed in argument.
        experiment_manager.interrupt_running_procedure()

    def _on_global_parameters_edited(self, parameters: ParameterNamespace) -> None:
        with self.session_maker() as session:
            session.set_global_parameters(parameters)
            logger.info(f"Global parameters written to storage: {parameters}")
        self.sequence_widget.set_available_parameter_names(parameters.names())

    def signal_exception_while_running_sequence(self, exception: Exception):
        # This is a bit ugly because on_procedure_exception runs a dialog, which
        # messes up the event loop, so instead we schedule the exception handling
        # to be done in the next event loop iteration.
        self.timer.singleShot(
            0, functools.partial(self.on_procedure_exception, exception)
        )



"""
The MIT License (MIT)

Copyright (c) 2012-2014 Alexander Turkin
Copyright (c) 2014 William Hallatt
Copyright (c) 2015 Jacob Dawid
Copyright (c) 2016 Luca Weiss
Copyright (c) 2017 fbjorn

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# This is a copy of https://github.com/fbjorn/QtWaitingSpinner/blob/master/pyqtspinner/spinner.py
# updated from PyQt5 to PySide6

import math

from PySide6.QtCore import Qt, QTimer, QRect
from PySide6.QtGui import QColor, QPaintEvent, QPainter
from PySide6.QtWidgets import QWidget


# pylint: disable=too-many-instance-attributes,too-many-arguments
class WaitingSpinner(QWidget):
    """WaitingSpinner is a highly configurable, custom spinner widget."""

    def __init__(
        self,
        parent: QWidget,
        center_on_parent: bool = True,
        disable_parent_when_spinning: bool = False,
        modality: Qt.WindowModality = Qt.WindowModality.NonModal,
        roundness: float = 100.0,
        fade: float = 80.0,
        lines: int = 20,
        line_length: int = 10,
        line_width: int = 2,
        radius: int = 10,
        speed: float = math.pi / 2,
        color: QColor = QColor(0, 0, 0),
    ) -> None:
        super().__init__(parent)

        self._center_on_parent: bool = center_on_parent
        self._disable_parent_when_spinning: bool = disable_parent_when_spinning

        self._color: QColor = color
        self._roundness: float = roundness
        self._minimum_trail_opacity: float = math.pi
        self._trail_fade_percentage: float = fade
        self._revolutions_per_second: float = speed
        self._number_of_lines: int = lines
        self._line_length: int = line_length
        self._line_width: int = line_width
        self._inner_radius: int = radius
        self._current_counter: int = 0
        self._is_spinning: bool = False

        self._timer: QTimer = QTimer(self)
        self._timer.timeout.connect(self._rotate)
        self._update_size()
        self._update_timer()
        self.hide()

        self.setWindowModality(modality)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

    def paintEvent(self, _: QPaintEvent) -> None:  # pylint: disable=invalid-name
        """Paint the WaitingSpinner."""
        self._update_position()
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.GlobalColor.transparent)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)

        if self._current_counter >= self._number_of_lines:
            self._current_counter = 0

        painter.setPen(Qt.PenStyle.NoPen)
        for i in range(self._number_of_lines):
            painter.save()
            painter.translate(
                self._inner_radius + self._line_length,
                self._inner_radius + self._line_length,
            )
            rotate_angle = 360 * i / self._number_of_lines
            painter.rotate(rotate_angle)
            painter.translate(self._inner_radius, 0)
            distance = self._line_count_distance_from_primary(
                i, self._current_counter, self._number_of_lines
            )
            color = self._current_line_color(
                distance,
                self._number_of_lines,
                self._trail_fade_percentage,
                self._minimum_trail_opacity,
                self._color,
            )
            painter.setBrush(color)
            painter.drawRoundedRect(
                QRect(
                    0,
                    -self._line_width // 2,
                    self._line_length,
                    self._line_width,
                ),
                self._roundness,
                self._roundness,
                Qt.SizeMode.RelativeSize,
            )
            painter.restore()

    def start(self) -> None:
        """Show and start spinning the WaitingSpinner."""
        self._update_position()
        self._is_spinning = True
        self.show()

        if self.parentWidget and self._disable_parent_when_spinning:
            self.parentWidget().setEnabled(False)

        if not self._timer.isActive():
            self._timer.start()
            self._current_counter = 0

    def stop(self) -> None:
        """Hide and stop spinning the WaitingSpinner."""
        self._is_spinning = False
        self.hide()

        if self.parentWidget() and self._disable_parent_when_spinning:
            self.parentWidget().setEnabled(True)

        if self._timer.isActive():
            self._timer.stop()
            self._current_counter = 0

    @property
    def color(self) -> QColor:
        """Return color of WaitingSpinner."""
        return self._color

    @color.setter
    def color(self, color: Qt.GlobalColor = Qt.GlobalColor.black) -> None:
        """Set color of WaitingSpinner."""
        self._color = QColor(color)

    @property
    def roundness(self) -> float:
        """Return roundness of WaitingSpinner."""
        return self._roundness

    @roundness.setter
    def roundness(self, roundness: float) -> None:
        """Set color of WaitingSpinner."""
        self._roundness = max(0.0, min(100.0, roundness))

    @property
    def minimum_trail_opacity(self) -> float:
        """Return minimum trail opacity of WaitingSpinner."""
        return self._minimum_trail_opacity

    @minimum_trail_opacity.setter
    def minimum_trail_opacity(self, minimum_trail_opacity: float) -> None:
        """Set minimum trail opacity of WaitingSpinner."""
        self._minimum_trail_opacity = minimum_trail_opacity

    @property
    def trail_fade_percentage(self) -> float:
        """Return trail fade percentage of WaitingSpinner."""
        return self._trail_fade_percentage

    @trail_fade_percentage.setter
    def trail_fade_percentage(self, trail: float) -> None:
        """Set trail fade percentage of WaitingSpinner."""
        self._trail_fade_percentage = trail

    @property
    def revolutions_per_second(self) -> float:
        """Return revolutions per second of WaitingSpinner."""
        return self._revolutions_per_second

    @revolutions_per_second.setter
    def revolutions_per_second(self, revolutions_per_second: float) -> None:
        """Set revolutions per second of WaitingSpinner."""
        self._revolutions_per_second = revolutions_per_second
        self._update_timer()

    @property
    def number_of_lines(self) -> int:
        """Return number of lines of WaitingSpinner."""
        return self._number_of_lines

    @number_of_lines.setter
    def number_of_lines(self, lines: int) -> None:
        """Set number of lines of WaitingSpinner."""
        self._number_of_lines = lines
        self._current_counter = 0
        self._update_timer()

    @property
    def line_length(self) -> int:
        """Return line length of WaitingSpinner."""
        return self._line_length

    @line_length.setter
    def line_length(self, length: int) -> None:
        """Set line length of WaitingSpinner."""
        self._line_length = length
        self._update_size()

    @property
    def line_width(self) -> int:
        """Return line width of WaitingSpinner."""
        return self._line_width

    @line_width.setter
    def line_width(self, width: int) -> None:
        """Set line width of WaitingSpinner."""
        self._line_width = width
        self._update_size()

    @property
    def inner_radius(self) -> int:
        """Return inner radius size of WaitingSpinner."""
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, radius: int) -> None:
        """Set inner radius size of WaitingSpinner."""
        self._inner_radius = radius
        self._update_size()

    @property
    def is_spinning(self) -> bool:
        """Return actual spinning status of WaitingSpinner."""
        return self._is_spinning

    def _rotate(self) -> None:
        """Rotate the WaitingSpinner."""
        self._current_counter += 1
        if self._current_counter >= self._number_of_lines:
            self._current_counter = 0
        self.update()

    def _update_size(self) -> None:
        """Update the size of the WaitingSpinner."""
        size = (self._inner_radius + self._line_length) * 2
        self.setFixedSize(size, size)

    def _update_timer(self) -> None:
        """Update the spinning speed of the WaitingSpinner."""
        self._timer.setInterval(
            int(1000 / (self._number_of_lines * self._revolutions_per_second))
        )

    def _update_position(self) -> None:
        """Center WaitingSpinner on parent widget."""
        if self.parentWidget() and self._center_on_parent:
            self.move(
                (self.parentWidget().width() - self.width()) // 2,
                (self.parentWidget().height() - self.height()) // 2,
            )

    @staticmethod
    def _line_count_distance_from_primary(
        current: int, primary: int, total_nr_of_lines: int
    ) -> int:
        """Return the amount of lines from _current_counter."""
        distance = primary - current
        if distance < 0:
            distance += total_nr_of_lines
        return distance

    @staticmethod
    def _current_line_color(
        count_distance: int,
        total_nr_of_lines: int,
        trail_fade_perc: float,
        min_opacity: float,
        color_input: QColor,
    ) -> QColor:
        """Returns the current color for the WaitingSpinner."""
        color = QColor(color_input)
        if count_distance == 0:
            return color
        min_alpha_f = min_opacity / 100.0
        distance_threshold = int(
            math.ceil((total_nr_of_lines - 1) * trail_fade_perc / 100.0)
        )
        if count_distance > distance_threshold:
            color.setAlphaF(min_alpha_f)
        else:
            alpha_diff = color.alphaF() - min_alpha_f
            gradient = alpha_diff / float(distance_threshold + 1)
            result_alpha = color.alphaF() - gradient * count_distance
            # If alpha is out of bounds, clip it.
            result_alpha = min(1.0, max(0.0, result_alpha))
            color.setAlphaF(result_alpha)
        return color



from __future__ import annotations

import collections
import time
from collections.abc import Set, Mapping
from typing import Any, Optional

import anyio
import attrs

from caqtus.device import DeviceName, DeviceController
from caqtus.device.remote import DeviceProxy
from caqtus.types.data import DataLabel, Data
from .._logger import logger
from ..shot_timing import ShotTimer


@attrs.define
class DeviceRunConfig:
    device: DeviceProxy
    controller_type: type[DeviceController]
    parameters: Mapping[str, Any]


@attrs.define
class DeviceRunInfo:
    controller: DeviceController
    device: DeviceProxy
    parameters: Mapping[str, Any]


class ShotEventDispatcher:
    def __init__(self, device_run_configs: Mapping[DeviceName, DeviceRunConfig]):
        self._device_infos: dict[DeviceName, DeviceRunInfo] = {
            name: DeviceRunInfo(
                controller=config.controller_type(name, self),
                device=config.device,
                parameters=config.parameters,
            )
            for name, config in device_run_configs.items()
        }

        self._controllers: dict[DeviceName, DeviceController] = {
            name: info.controller for name, info in self._device_infos.items()
        }

        self._acquisition_events: dict[DataLabel, anyio.Event] = (
            collections.defaultdict(anyio.Event)
        )
        self._acquired_data: dict[DataLabel, Data] = {}
        self._start_time = 0.0

        self._shot_timer: Optional[ShotTimer] = None

    async def run_shot(self, timeout: float) -> Mapping[DataLabel, Data]:
        try:
            return await self._run_shot(timeout)
        except:
            stats = {
                name: controller._debug_stats()
                for name, controller in self._controllers.items()
            }
            logger.debug("Shot trace: %s", stats)
            raise

    async def _run_shot(self, timeout: float) -> Mapping[DataLabel, Data]:
        self._start_time = time.monotonic()
        result = {}
        with anyio.fail_after(timeout):
            async with anyio.create_task_group() as tg:
                for name, info in self._device_infos.items():
                    # noinspection PyProtectedMember
                    tg.start_soon(
                        _save_in_dict,
                        info.controller._run_shot(info.device, **info.parameters),
                        name,
                        result,
                    )

        data = self.acquired_data()
        data[DataLabel("shot analytics")] = result
        return data

    def shot_time(self) -> float:
        return time.monotonic() - self._start_time

    async def wait_all_devices_ready(self) -> ShotTimer:
        async with anyio.create_task_group() as tg:
            for controller in self._controllers.values():
                if not controller._signaled_ready.is_set():
                    # noinspection PyProtectedMember
                    tg.start_soon(controller._signaled_ready.wait)

        if self._shot_timer is None:
            # noinspection PyProtectedMember
            self._shot_timer = ShotTimer._create()
        return self._shot_timer

    async def wait_data_acquired(
        self, waiting_device: DeviceName, label: DataLabel
    ) -> Data:
        if label not in self._acquired_data:
            await self._acquisition_events[label].wait()
        return self._acquired_data[label]

    def signal_data_acquired(
        self, emitting_device: DeviceName, label: DataLabel, data: Data
    ) -> None:
        if label in self._acquired_data:
            raise KeyError(f"There is already data acquired for label {label}")
        else:
            self._acquired_data[label] = data
            if label in self._acquisition_events:
                self._acquisition_events[label].set()

    def waiting_on_data(self) -> Set[DataLabel]:
        return set(
            label
            for label, event in self._acquisition_events.items()
            if not event.is_set()
        )

    def acquired_data(self) -> dict[DataLabel, Data]:
        if not_acquired := self.waiting_on_data():
            raise RuntimeError(
                f"Still waiting on data acquisition for labels {not_acquired}"
            )
        return self._acquired_data


async def _save_in_dict(task, key, dictionary):
    dictionary[key] = await task



from __future__ import annotations

import functools
from collections.abc import Callable
from typing import TypeVar

import attrs
from typing_extensions import Protocol

from caqtus.device import DeviceConfiguration
from caqtus.device.configuration.serializer import (
    DeviceConfigJSONSerializer,
    DeviceConfigJSONSerializerProtocol,
)
from caqtus.types.iteration import (
    IterationConfiguration,
    StepsConfiguration,
)
from caqtus.types.timelane.serializer import (
    TimeLaneSerializerProtocol,
    TimeLaneSerializer,
)
from caqtus.utils import serialization
from caqtus.utils.serialization import JSON
from ..shot import TimeLane
from ...types.timelane import TimeLanes

T = TypeVar("T", bound=DeviceConfiguration)


class SerializerProtocol(Protocol):
    sequence_serializer: SequenceSerializer
    device_configuration_serializer: DeviceConfigJSONSerializerProtocol
    time_lane_serializer: TimeLaneSerializerProtocol

    def dump_device_configuration(
        self, config: DeviceConfiguration
    ) -> tuple[str, serialization.JSON]:
        return self.device_configuration_serializer.dump_device_configuration(config)

    def load_device_configuration(
        self, tag: str, content: serialization.JSON
    ) -> DeviceConfiguration:
        return self.device_configuration_serializer.load_device_configuration(
            tag, content
        )

    def construct_sequence_iteration(
        self, content: serialization.JSON
    ) -> IterationConfiguration:
        return self.sequence_serializer.iteration_constructor(content)

    def dump_sequence_iteration(
        self, iteration: IterationConfiguration
    ) -> serialization.JSON:
        return self.sequence_serializer.iteration_serializer(iteration)

    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> serialization.JSON:
        return self.time_lane_serializer.unstructure_time_lanes(time_lanes)

    def structure_time_lanes(self, content: serialization.JSON) -> TimeLanes:
        return self.time_lane_serializer.structure_time_lanes(content)

    def construct_time_lane(self, content: serialization.JSON) -> TimeLane:
        return self.time_lane_serializer.load(content)


@attrs.define
class Serializer(SerializerProtocol):
    """Serialize and deserialize user objects."""

    sequence_serializer: SequenceSerializer
    device_configuration_serializer: DeviceConfigJSONSerializer
    time_lane_serializer: TimeLaneSerializer

    @classmethod
    def default(cls) -> Serializer:
        return Serializer(
            sequence_serializer=default_sequence_serializer,
            device_configuration_serializer=DeviceConfigJSONSerializer(),
            time_lane_serializer=TimeLaneSerializer(),
        )

    def register_device_configuration(
        self,
        config_type: type[T],
        dumper: Callable[[T], JSON],
        constructor: Callable[[JSON], T],
    ) -> None:
        self.device_configuration_serializer.register_device_configuration(
            config_type, dumper, constructor
        )

    def register_iteration_configuration_serializer(
        self,
        dumper: Callable[[IterationConfiguration], serialization.JSON],
        constructor: Callable[[serialization.JSON], IterationConfiguration],
    ) -> None:
        self.sequence_serializer = SequenceSerializer(
            iteration_serializer=dumper,
            iteration_constructor=constructor,
        )


@attrs.frozen
class SequenceSerializer:
    iteration_serializer: Callable[[IterationConfiguration], serialization.JSON]
    iteration_constructor: Callable[[serialization.JSON], IterationConfiguration]


@functools.singledispatch
def default_iteration_configuration_serializer(
    iteration_configuration: IterationConfiguration,
) -> serialization.JSON:
    raise TypeError(
        f"Cannot serialize iteration configuration of type "
        f"{type(iteration_configuration)}"
    )


@default_iteration_configuration_serializer.register
def _(
    iteration_configuration: StepsConfiguration,
):
    content = serialization.converters["json"].unstructure(iteration_configuration)
    content["type"] = "steps"
    return content


def default_iteration_configuration_constructor(
    iteration_content: serialization.JSON,
) -> IterationConfiguration:
    iteration_type = iteration_content.pop("type")
    if iteration_type == "steps":
        return serialization.converters["json"].structure(
            iteration_content, StepsConfiguration
        )
    else:
        raise ValueError(f"Unknown iteration type {iteration_type}")


default_sequence_serializer = SequenceSerializer(
    iteration_serializer=default_iteration_configuration_serializer,
    iteration_constructor=default_iteration_configuration_constructor,
)



from collections.abc import Mapping, Callable, Iterable
from typing import Protocol, Any, overload

from caqtus.types.variable_name import DottedVariableName


class Evaluable[T](Protocol):
    """Defines an object that can be evaluated to a value."""

    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> T:
        """Evaluate the object using the given variables.

        Args:
            variables: A mapping of variable names to values.
                The evaluable object will be interpreted given the values of these
                variables.

        Returns:
            The evaluated value.
        """

        raise NotImplementedError


@overload
def evaluate[
    T
](evaluable: Evaluable[T], variables: Mapping[DottedVariableName, Any],) -> T: ...


@overload
def evaluate[
    T, R
](
    evaluable: Evaluable[T],
    variables: Mapping[DottedVariableName, Any],
    transformer: Callable[[T], R],
) -> R: ...


@overload
def evaluate[
    T, R1, R2
](
    evaluable: Evaluable[T],
    variables: Mapping[DottedVariableName, Any],
    transformer: tuple[Callable[[T], R1], Callable[[R1], R2]],
) -> R2: ...


@overload
def evaluate[
    T
](
    evaluable: Evaluable[T],
    variables: Mapping[DottedVariableName, Any],
    transformer: tuple[Callable, ...],
) -> Any: ...


def evaluate(evaluable, variables, transformer=None):
    """Evaluate the given evaluable object using the given variables.

    Args:
        evaluable: An object that can be evaluated to a value.
        variables: A mapping of variable names to values.
            The evaluable object will be interpreted given the values of these
            variables.
        transformer: Functions that transforms the evaluated value.
            For example, this can be used to convert the value to a specific unit or
            ensure that it is within a certain range.
            If None, the evaluated value is returned as is.
    """

    value = evaluable.evaluate(variables)

    if transformer is None:
        return value

    if callable(transformer):
        return transformer(value)

    if not isinstance(transformer, Iterable):
        raise TypeError(
            "The transformer must be a callable or an iterable of callables."
        )
    for t in transformer:
        value = t(value)

    return value



from .analog_value import (
    AnalogValue,
    NotAnalogValueError,
    is_analog_value,
    is_quantity,
    add_unit,
    get_unit,
    magnitude_in_unit,
)
from .parameter import Parameter, is_parameter
from .parameter_namespace import ParameterNamespace

__all__ = [
    "AnalogValue",
    "NotAnalogValueError",
    "add_unit",
    "is_analog_value",
    "Parameter",
    "is_parameter",
    "ParameterNamespace",
    "is_quantity",
    "get_unit",
    "magnitude_in_unit",
]



import abc
from contextlib import ExitStack, AbstractContextManager
from typing import Optional, TypeVar, Self

import attrs

from ._device import Device
from .._name import DeviceName

_T = TypeVar("_T")


@attrs.define
class RuntimeDevice(Device, abc.ABC):
    """An implementation of the Device class that provides some useful operations.

    Class inheriting from RuntimeDevice can use the methods `_add_closing_callback` and
    `_enter_context` to facilitate managing resources.

    Fields:
        name: A unique name given to the device. Cannot be changed during the lifetime
        of the device.

    Warnings:
        All device classes subclassed from this class should call parent_item class
        initialize and close methods when overwriting them.
    """

    name: DeviceName = attrs.field(on_setattr=attrs.setters.frozen)

    _close_stack: Optional[ExitStack] = attrs.field(init=False, default=None)

    @name.validator  # type: ignore
    def _validate_name(self, _, value):
        if not isinstance(value, str):
            raise TypeError(f"Expected DeviceName, got {type(value)}")

    @abc.abstractmethod
    def initialize(self) -> None:
        """Initiate the communication to the device.

        This method is meant to be reimplemented for each specific device.
        The base class implementation registers the device in the list of devices
        already in use.
        It must be called when subclassing this class.
        """

        pass

    def _add_closing_callback(self, callback, /, *args, **kwargs):
        """Add a callback function to be called when the device is closed.

        Callbacks will be called in the reverse order they were added.
        The callback is called with the same arguments as passed to this method.
        """

        if self._close_stack is None:
            raise UninitializedDeviceError(
                f"Method RuntimeDevice.initialize must be called on the instance "
                f"before adding shutdown callbacks."
            )
        self._close_stack.callback(callback, *args, **kwargs)

    def _enter_context(self, cm: AbstractContextManager[_T]) -> _T:
        """Enter a context manager to be closed when the device is closed."""

        if self._close_stack is None:
            raise UninitializedDeviceError(
                f"Method RuntimeDevice.initialize must be called on the instance "
                f"before entering context managers."
            )
        return self._close_stack.enter_context(cm)

    def close(self) -> None:
        """Close the communication to the device and free the resources used.

        This method must be called once when use of the device is finished.
        The base class implementation unwinds the stack of closing callbacks that where
        registered when it is called.
        If you only use `_enter_context` and `_add_closing_callback`, there is no need
        to reimplement this method in subclasses.
        """

        if self._close_stack is None:
            raise UninitializedDeviceError(
                f"method RuntimeDevice for {self.name} must be entered called before calling close."
            )
        self._close_stack.close()
        self._close_stack = None

    def __enter__(self) -> Self:
        """Initialize the device.

        When entering the device as a context manager, it will try to acquire necessary
        resources by calling `initialize`.
        If an error occurs while calling initialization, the `close` method will be
        called.
        Typically, subclasses only need to reimplement `initialize` using only
        `_enter_context` and `_add_closing_callback`.
        This will ensure proper cleanup.
        """

        self._close_stack = ExitStack()

        try:
            self.initialize()
        except:
            self.close()
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get_name(self) -> DeviceName:
        return self.name


class UninitializedDeviceError(Exception):
    """Raised when a device is used before being initialized"""

    pass



from typing import Protocol

from caqtus.device.configuration.serializer import DeviceConfigJSONSerializerProtocol
from caqtus.experiment_control.device_manager_extension import (
    DeviceManagerExtensionProtocol,
)
from caqtus.gui.condetrol._extension import CondetrolExtensionProtocol
from caqtus.types.timelane.serializer import TimeLaneSerializerProtocol


class CaqtusExtensionProtocol(Protocol):
    condetrol_extension: CondetrolExtensionProtocol
    device_configurations_serializer: DeviceConfigJSONSerializerProtocol
    time_lane_serializer: TimeLaneSerializerProtocol
    device_manager_extension: DeviceManagerExtensionProtocol



from caqtus.types.timelane import (
    TimeLane,
    TimeLanes,
    DigitalTimeLane,
    AnalogTimeLane,
    Ramp,
    CameraTimeLane,
    TakePicture,
)
from ..sequence.shot import Shot

__all__ = [
    "TimeLane",
    "TimeLanes",
    "DigitalTimeLane",
    "AnalogTimeLane",
    "Ramp",
    "CameraTimeLane",
    "TakePicture",
    "Shot",
]



import abc
from collections.abc import Callable, Set
from typing import TypeVar, Generic, TypeAlias

import caqtus.gui.qtutil.qabc as qabc
from PySide6.QtCore import Signal
from caqtus.types.iteration import IterationConfiguration
from caqtus.types.variable_name import DottedVariableName

T = TypeVar("T", bound=IterationConfiguration)


class SequenceIterationEditor(Generic[T], metaclass=qabc.QABCMeta):
    iteration_edited = Signal(IterationConfiguration)

    @abc.abstractmethod
    def get_iteration(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def set_iteration(self, iteration: T):
        raise NotImplementedError

    @abc.abstractmethod
    def set_read_only(self, read_only: bool):
        raise NotImplementedError

    @abc.abstractmethod
    def set_available_parameter_names(self, parameter_names: Set[DottedVariableName]):
        """Set the names of the parameters that are defined externally."""

        raise NotImplementedError


IterationEditorCreator: TypeAlias = Callable[[T], SequenceIterationEditor[T]]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'single_shot_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QAction,
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QDockWidget,
    QHeaderView,
    QMainWindow,
    QMdiArea,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QTreeView,
    QVBoxLayout,
    QWidget,
)


class Ui_SingleShotWidget(object):
    def setupUi(self, SingleShotWidget):
        if not SingleShotWidget.objectName():
            SingleShotWidget.setObjectName("SingleShotWidget")
        SingleShotWidget.resize(800, 600)
        self._action_cascade = QAction(SingleShotWidget)
        self._action_cascade.setObjectName("_action_cascade")
        self._action_tile = QAction(SingleShotWidget)
        self._action_tile.setObjectName("_action_tile")
        self.actionImage = QAction(SingleShotWidget)
        self.actionImage.setObjectName("actionImage")
        self.actionParameters = QAction(SingleShotWidget)
        self.actionParameters.setObjectName("actionParameters")
        self.actionAtoms = QAction(SingleShotWidget)
        self.actionAtoms.setObjectName("actionAtoms")
        self.actionSave = QAction(SingleShotWidget)
        self.actionSave.setObjectName("actionSave")
        self.actionSave_as = QAction(SingleShotWidget)
        self.actionSave_as.setObjectName("actionSave_as")
        self.actionLoad = QAction(SingleShotWidget)
        self.actionLoad.setObjectName("actionLoad")
        self.centralwidget = QWidget(SingleShotWidget)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self._mdi_area = QMdiArea(self.centralwidget)
        self._mdi_area.setObjectName("_mdi_area")

        self.verticalLayout.addWidget(self._mdi_area)

        SingleShotWidget.setCentralWidget(self.centralwidget)
        self.menuBar = QMenuBar(SingleShotWidget)
        self.menuBar.setObjectName("menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 800, 21))
        self.menuWindow = QMenu(self.menuBar)
        self.menuWindow.setObjectName("menuWindow")
        self.menu_add_viewer = QMenu(self.menuBar)
        self.menu_add_viewer.setObjectName("menu_add_viewer")
        self.menuWorkspace = QMenu(self.menuBar)
        self.menuWorkspace.setObjectName("menuWorkspace")
        SingleShotWidget.setMenuBar(self.menuBar)
        self._shot_selector_dock = QDockWidget(SingleShotWidget)
        self._shot_selector_dock.setObjectName("_shot_selector_dock")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self._shot_selector_dock.sizePolicy().hasHeightForWidth()
        )
        self._shot_selector_dock.setSizePolicy(sizePolicy)
        self._shot_selector_dock.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self._shot_selector_dock.setWidget(self.dockWidgetContents)
        SingleShotWidget.addDockWidget(Qt.TopDockWidgetArea, self._shot_selector_dock)
        self._sequence_hierarchy_dock = QDockWidget(SingleShotWidget)
        self._sequence_hierarchy_dock.setObjectName("_sequence_hierarchy_dock")
        self._sequence_hierarchy_dock.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.dockWidgetContents_2 = QWidget()
        self.dockWidgetContents_2.setObjectName("dockWidgetContents_2")
        self.verticalLayout_2 = QVBoxLayout(self.dockWidgetContents_2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self._sequence_hierarchy_view = QTreeView(self.dockWidgetContents_2)
        self._sequence_hierarchy_view.setObjectName("_sequence_hierarchy_view")

        self.verticalLayout_2.addWidget(self._sequence_hierarchy_view)

        self._sequence_hierarchy_dock.setWidget(self.dockWidgetContents_2)
        SingleShotWidget.addDockWidget(
            Qt.LeftDockWidgetArea, self._sequence_hierarchy_dock
        )

        self.menuBar.addAction(self.menuWorkspace.menuAction())
        self.menuBar.addAction(self.menu_add_viewer.menuAction())
        self.menuBar.addAction(self.menuWindow.menuAction())
        self.menuWindow.addAction(self._action_cascade)
        self.menuWindow.addAction(self._action_tile)
        self.menuWorkspace.addAction(self.actionSave_as)
        self.menuWorkspace.addAction(self.actionLoad)

        self.retranslateUi(SingleShotWidget)

        QMetaObject.connectSlotsByName(SingleShotWidget)

    # setupUi

    def retranslateUi(self, SingleShotWidget):
        SingleShotWidget.setWindowTitle(
            QCoreApplication.translate("SingleShotWidget", "MainWindow", None)
        )
        self._action_cascade.setText(
            QCoreApplication.translate("SingleShotWidget", "Cascade", None)
        )
        self._action_tile.setText(
            QCoreApplication.translate("SingleShotWidget", "Tile", None)
        )
        self.actionImage.setText(
            QCoreApplication.translate("SingleShotWidget", "Image", None)
        )
        self.actionParameters.setText(
            QCoreApplication.translate("SingleShotWidget", "Parameters", None)
        )
        self.actionAtoms.setText(
            QCoreApplication.translate("SingleShotWidget", "Atoms", None)
        )
        self.actionSave.setText(
            QCoreApplication.translate("SingleShotWidget", "Save", None)
        )
        self.actionSave_as.setText(
            QCoreApplication.translate("SingleShotWidget", "Save as", None)
        )
        self.actionLoad.setText(
            QCoreApplication.translate("SingleShotWidget", "Load", None)
        )
        self.menuWindow.setTitle(
            QCoreApplication.translate("SingleShotWidget", "Windows", None)
        )
        self.menu_add_viewer.setTitle(
            QCoreApplication.translate("SingleShotWidget", "Add viewer", None)
        )
        self.menuWorkspace.setTitle(
            QCoreApplication.translate("SingleShotWidget", "Workspace", None)
        )

    # retranslateUi



from PySide6.QtWidgets import QLineEdit


from caqtus.types.expression import Expression


class ExpressionWidget(QLineEdit):
    """A widget to enter an expression"""

    def set_expression(self, expression: Expression):
        self.setText(expression.body)

    def get_expression(self) -> Expression:
        return Expression(self.text())



from .combinable_importers import CombinableLoader
from .combinable_importers import join
from .load_parameters import LoadShotParameters
from .load_shot_id import LoadShotId
from .load_shot_info import LoadShotTime
from .shot_data import ShotData, ShotImporter, DataImporter

__all__ = [
    "CombinableLoader",
    "LoadShotParameters",
    "LoadShotId",
    "LoadShotTime",
    "ShotData",
    "ShotImporter",
    "join",
    "DataImporter",
]



from .async_session import AsyncExperimentSession
from .data_type import DataType
from .experiment_session import ExperimentSession
from .path import BoundSequencePath, PureSequencePath
from .path_hierarchy import PathHierarchy
from .sequence import Sequence, Shot
from .session_maker import ExperimentSessionMaker

__all__ = [
    "ExperimentSession",
    "ExperimentSessionMaker",
    "PureSequencePath",
    "BoundSequencePath",
    "DataType",
    "PathHierarchy",
    "Sequence",
    "Shot",
    "AsyncExperimentSession",
]



"""
This module contains functions and classes to compile user-friendly sequence
configurations into low-level device parameters.
"""

from caqtus.types.units.unit_namespace import units
from ._device_compiler import DeviceCompiler, DeviceNotUsedException
from .compilation_contexts import ShotContext, SequenceContext
from .variable_namespace import VariableNamespace

__all__ = [
    "VariableNamespace",
    "units",
    "ShotContext",
    "SequenceContext",
    "DeviceCompiler",
    "DeviceNotUsedException",
]



"""This module contains classes to configure the components to use for a given
experiment.

The class :class:`Experiment` can be used to register extensions for a specific setup.

The class :class:`DeviceExtension` can be used to specify a new type of device.

The class :class:`TimeLaneExtension` can be used to specify a new type of time lane.
"""

from ._experiment import Experiment
from .device_extension import DeviceExtension
from .time_lane_extension import TimeLaneExtension

__all__ = ["DeviceExtension", "TimeLaneExtension", "Experiment"]



from typing import Callable, Optional, ParamSpec, TypeVar

from PySide6.QtCore import QThread, Qt
from PySide6.QtWidgets import QWidget, QDialog, QMessageBox

P = ParamSpec("P")
T = TypeVar("T")


class WorkerThread(QThread):
    def __init__(self, function: Callable[[P], T], *args: P.args, **kwargs: P.kwargs):
        super().__init__()
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.exception = None
        self.result: Optional[T] = None

    def run(self):
        def _run():
            try:
                self.result = self.function(*self.args, **self.kwargs)
            except Exception as e:
                self.exception = e
            finally:
                self.finished.emit()

        return _run()


class BlockingWidget(QDialog):
    def __init__(self, msg: str, parent: QWidget):
        super().__init__(parent)
        self.setModal(True)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)


def run_with_wip_widget(
    parent: QWidget,
    msg: str,
    function: Callable[[P], T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    message_box = QMessageBox(parent)
    message_box.setText(msg)
    message_box.setWindowFlags(
        (message_box.windowFlags() | Qt.WindowType.Window.CustomizeWindowHint)
        & Qt.WindowType.Window.FramelessWindowHint
    )
    worker = WorkerThread(function, *args, **kwargs)
    worker.finished.connect(message_box.close)
    worker.start()
    message_box.exec()
    if worker.exception is not None:
        raise worker.exception
    return worker.result



from collections.abc import Mapping
from typing import Protocol, runtime_checkable, Optional

import attrs

from .device import Device


@attrs.frozen
class TestError:
    message: str = attrs.field(converter=str)


@attrs.frozen
class DataTestError:
    data: Mapping[str, float] = attrs.field(
        converter=dict,
        validator=attrs.validators.deep_mapping(
            key_validator=attrs.validators.instance_of(str),
            value_validator=attrs.validators.instance_of(float),
        ),
    )


@attrs.frozen
class TestErrorGroup(TestError):
    results: tuple[TestError, ...] = attrs.field(
        converter=tuple,
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(TestError)
        ),
    )


@runtime_checkable
class TestableDevice(Device, Protocol):
    """
    Interface that defines a device that can be tested to check it is working properly.

    Instances of this class have a `run_test` method that can be called whenever the device is not running to check that
    the device is working properly.
    """

    def run_test(self) -> Optional[TestError]:
        """Test if a device pass the test.

        It is strongly advised to do the test only by reading data from the device and without changing its state if it
        also a device controlling some channels on the experiment.

        Returns:
            None, if the device is working as expected or a test error containing extra information if the device is not
            working as it should.
        """

        ...


@runtime_checkable
class RecalibratingDevice(TestableDevice, Protocol):
    """Interface for a device that can recalibrate itself if it fails to pass its test."""

    def recalibrate(self, test_error: Optional[TestError]) -> bool:
        """Attempts to recalibrate the device.

        Args:
            test_error: An error that occurred previously. If None is passed, the calibration should perform all the
            actions necessary to put the device back in running order. If a specific test error is passed, the device
            can attempt to fix only this specific error instead of doing all the fix operations.

        Returns:
            True, if the recalibration was successful, False, if it failed.
        """

        ...



from typing import Optional

from PySide6.QtWidgets import QWidget, QSpinBox, QFormLayout

from caqtus.device.camera import CameraConfigurationType
from caqtus.utils.roi import RectangularROI, Width, Height
from .._device_configuration_editor import FormDeviceConfigurationEditor


class CameraConfigurationEditor(FormDeviceConfigurationEditor[CameraConfigurationType]):
    """A widget that allows to edit the configuration of a camera.

    This widget has the same fields as the base form editor and adds a field to edit
    the region of interest of the image to take.

    The widget is generic in the type :any:`CameraConfigurationType` to allow to use
    it with different camera configuration types.

    Args:
        configuration: The initial camera configuration to display in the editor.
        parent: The parent widget of the editor.

    Attributes:
        roi_editor: The widget that allows to edit the region of interest of the image.
    """

    def __init__(
        self, configuration: CameraConfigurationType, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(configuration, parent)

        roi = configuration.roi
        self.roi_editor = RectangularROIEditor(
            max_width=roi.original_width, max_height=roi.original_height, parent=self
        )
        self.append_row("ROI", self.roi_editor)
        self.roi_editor.set_roi(roi)

    def get_configuration(self) -> CameraConfigurationType:
        """Return a new camera configuration representing what is currently displayed.

        The configuration returns has the remote server and the ROI set from what is
        displayed in the editor.

        Other fields should be updated by subclasses.
        """

        configuration = super().get_configuration()
        configuration.roi = self.roi_editor.get_roi()
        return configuration


class RectangularROIEditor(QWidget):
    """A widget that allows to edit a rectangular region of interest.

    Args:
        max_width: The maximum width in pixels that the region of interest can have.
        max_height: The maximum height in pixels that the region of interest can have.
    """

    def __init__(
        self,
        max_width: int,
        max_height: int,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)

        self._max_width = max_width
        self._max_height = max_height

        layout = QFormLayout(self)
        layout.setContentsMargins(0, 3, 0, 0)
        self.setLayout(layout)

        self._x_spinbox = QSpinBox(self)
        layout.addRow("X", self._x_spinbox)
        self._x_spinbox.setRange(0, 0)
        self._x_spinbox.setValue(0)

        self._width_spinbox = QSpinBox(self)
        layout.addRow("Width", self._width_spinbox)
        self._width_spinbox.setRange(1, self._max_width)
        self._width_spinbox.setValue(self._max_width)

        self._y_spinbox = QSpinBox(self)
        layout.addRow("Y", self._y_spinbox)
        self._y_spinbox.setRange(0, 0)
        self._y_spinbox.setValue(0)

        self._height_spinbox = QSpinBox(self)
        layout.addRow("Height", self._height_spinbox)
        self._height_spinbox.setRange(1, self._max_height)
        self._height_spinbox.setValue(self._max_height)

        self._x_spinbox.valueChanged.connect(self._on_x_value_changed)
        self._y_spinbox.valueChanged.connect(self._on_y_value_changed)

        self._width_spinbox.valueChanged.connect(self._on_width_value_changed)
        self._height_spinbox.valueChanged.connect(self._on_height_value_changed)

    def set_roi(self, roi: RectangularROI) -> None:
        """Set the values to be displayed in the editor."""

        self._max_width = roi.original_width
        self._max_height = roi.original_height

        # We first set x and y coordinates to 0 to have the full allowed range for
        # width and height spinboxes, otherwise the range would be limited by the
        # current x and y values.

        self._x_spinbox.setValue(0)
        self._width_spinbox.setValue(roi.width)
        self._x_spinbox.setValue(roi.x)

        self._y_spinbox.setValue(0)
        self._height_spinbox.setValue(roi.height)
        self._y_spinbox.setValue(roi.y)

    def get_roi(self) -> RectangularROI:
        """Return the values of the ROI currently displayed in the editor."""

        return RectangularROI(
            x=self._x_spinbox.value(),
            y=self._y_spinbox.value(),
            width=self._width_spinbox.value(),
            height=self._height_spinbox.value(),
            original_image_size=(Width(self._max_width), Height(self._max_height)),
        )

    def _on_x_value_changed(self, x: int) -> None:
        self._width_spinbox.setRange(1, self._max_width - x)

    def _on_y_value_changed(self, y: int) -> None:
        self._height_spinbox.setRange(1, self._max_height - y)

    def _on_width_value_changed(self, width: int) -> None:
        self._x_spinbox.setRange(0, self._max_width - width)

    def _on_height_value_changed(self, height: int) -> None:
        self._y_spinbox.setRange(0, self._max_height - height)



from typing import TypeGuard, Any

import numpy as np

from .data_type import Data


def is_data(data: Any) -> TypeGuard[Data]:
    """Check if data has a valid data type."""

    if isinstance(data, (bool, int, str, float, np.ndarray)):
        return True
    elif isinstance(data, (list, tuple)):
        return all(is_data(x) for x in data)
    elif isinstance(data, dict):
        return all(isinstance(k, str) and is_data(v) for k, v in data.items())

    return False



from tqdm import tqdm

from caqtus.session import ExperimentSession, PureSequencePath
from caqtus.session._return_or_raise import unwrap
from caqtus.session.sequence import State


def copy_sequence(
    source_session: ExperimentSession,
    target_session: ExperimentSession,
    sequence_path: PureSequencePath,
) -> None:
    """Copy a sequence from one session to another.

    The two sessions must be activated and must be different.
    """

    created_paths = unwrap(target_session.paths.create_path(sequence_path))
    for path in created_paths:
        target_session.paths.update_creation_date(
            path, unwrap(source_session.paths.get_path_creation_date(sequence_path))
        )
    target_session.sequences.create(
        sequence_path,
        source_session.sequences.get_parameters(sequence_path),
        source_session.sequences.get_iteration_configuration(sequence_path),
        source_session.sequences.get_time_lanes(sequence_path),
    )
    state = unwrap(source_session.sequences.get_state(sequence_path))
    if not state.is_editable():
        target_session.sequences.set_state(sequence_path, State.PREPARING)
        device_configs = source_session.sequences.get_device_configurations(
            sequence_path
        )
        target_session.sequences.set_device_configurations(
            sequence_path, device_configs
        )
        target_session.sequences.set_state(sequence_path, State.RUNNING)
        shots = unwrap(source_session.sequences.get_shots(sequence_path))
        for shot in tqdm(shots, desc=f"Copying {sequence_path}"):
            target_session.sequences.create_shot(
                sequence_path,
                shot_index=shot.index,
                shot_parameters=shot.get_parameters(source_session),
                shot_data=shot.get_data(source_session),
                shot_start_time=shot.get_start_time(source_session),
                shot_end_time=shot.get_end_time(source_session),
            )
        target_session.sequences.set_state(sequence_path, state)
    stats = unwrap(source_session.sequences.get_stats(sequence_path))
    target_session.sequences.update_start_and_end_time(
        sequence_path, stats.start_time, stats.stop_time
    )



import enum


class DataType(enum.Enum):
    """Enumeration for the type of data attached to a shot.

    Each value stored in the database is tagged with a data type. This represents how
    the data was generated for this shot and how it should be interpreted.

    Attributes:
        PARAMETER: The data is a value that was used to configure a shot. A shot can be
            understood as a function that takes a set of parameters and returns a set of
            measurements.
        MEASURE: The data is a measurement that was recorded during a shot. A shot can
            acquire multiple measurements such as the fluorescence of the MOT, an image,
            the presence of an atom, etc.
        ANALYSIS: The data is a value that was computed from one or more measurements.
        SCORE: This is used for optimization. It indicates the value of a score function
            for a given set of parameters.
    """

    PARAMETER = "parameter"
    MEASURE = "measure"
    ANALYSIS = "analysis"
    SCORE = "score"



import re

from PySide6.QtGui import Qt, QPalette
from PySide6.QtWidgets import QTreeWidgetItem, QApplication


def create_exception_tree(
    exception: BaseException, prepend: str = "error:"
) -> list[QTreeWidgetItem]:
    result = []
    text = str(exception)
    if isinstance(exception, ExceptionGroup):
        text = exception.args[0]
    text = process_text(text)
    exception_label = type(exception).__name__
    exception_item = QTreeWidgetItem(None, [prepend, exception_label, text])
    error_color = Qt.GlobalColor.red
    highlight_color = QApplication.palette().color(QPalette.ColorRole.Accent)
    exception_item.setForeground(0, highlight_color)
    exception_item.setForeground(1, error_color)
    result.append(exception_item)
    if hasattr(exception, "__notes__"):
        for note in exception.__notes__:
            note_item = QTreeWidgetItem(exception_item, ["", "", note])
            exception_item.addChild(note_item)
    if isinstance(exception, ExceptionGroup):
        for i, child_exception in enumerate(exception.exceptions):
            exception_item.addChildren(
                create_exception_tree(child_exception, f"Sub-error {i}")
            )
    if exception.__cause__ is not None:
        for cause in create_exception_tree(exception.__cause__, "because:"):
            exception_item.addChild(cause)
    return result


def process_text(text: str) -> str:
    color = QApplication.palette().color(QPalette.ColorRole.Accent).name()
    text = highlight_device_name(text, color)
    text = highlight_device_servers(text, color)
    text = highlight_expression(text, color)
    # device parameter must be highlighted before parameter, because they match the
    # same pattern
    text = hightlight_device_parameter(text, color)
    text = highlight_parameter(text, color)
    text = highlight_type(text, color)
    text = highlight_step(text, color)

    return text


def highlight_device_name(value: str, color) -> str:
    def replace(match):
        return f'device <b><font color="{color}">{match.group(1)}</font></b>'

    return re.sub(r"device '(.+?)'", replace, value)


def highlight_device_servers(text: str, color) -> str:
    def replace(match):
        return f'device server <b><font color="{color}">{match.group(1)}</font></b>'

    return re.sub(r"device server '(.+?)'", replace, text)


def highlight_expression(text: str, color) -> str:
    def replace(match):
        return f'expression <b><font color="{color}">{match.group(1)}</font></b>'

    return re.sub(r"expression '(.+?)'", replace, text)


def highlight_parameter(text: str, color) -> str:
    def replace(match):
        return f'parameter <b><font color="{color}">{match.group(1)}</font></b>'

    return re.sub(r"parameter '(.+?)'", replace, text)


def highlight_type(text: str, color) -> str:
    def replace(match):
        return f'type <b><font color="{color}">{match.group(1)}</font></b>'

    return re.sub(r"type '(.+?)'", replace, text)


def highlight_step(text: str, color) -> str:
    def replace(match):
        return f'step <b><font color="{color}">{match.group(1)} (<i>{match.group(2)}</i>)</font></b>'

    return re.sub(r"step ([0-9]+) \((.+?)\)", replace, text)


def hightlight_device_parameter(text: str, color) -> str:
    def replace(match):
        return f'device parameter <b><font color="{color}">{match.group(1)} = {match.group(2)}</font></b>'

    return re.sub(r"device parameter '(.+?)' = '(.+?)'", replace, text)



from __future__ import annotations

import contextlib
import copy
import logging
import threading
from collections.abc import Mapping, AsyncGenerator, AsyncIterable
from typing import Optional

import anyio
import anyio.to_process
import anyio.to_thread

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.session import ExperimentSessionMaker, PureSequencePath
from caqtus.session.sequence import State
from caqtus.shot_compilation import (
    DeviceCompiler,
    SequenceContext,
)
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.recoverable_exceptions import SequenceInterruptedException
from .shots_manager import ShotManager, ShotData, ShotScheduler
from .shots_manager import ShotRetryConfig
from ..device_manager_extension import DeviceManagerExtensionProtocol
from ..initialize_resources import create_shot_runner, create_shot_compiler

logger = logging.getLogger(__name__)


class SequenceManager:
    """Manages the execution of a sequence.

    Args:
        sequence: The sequence to run.
        session_maker: A factory for creating experiment sessions.
        This is used to connect to the storage in which to find the sequence.
        interruption_event: An event that is set to interrupt the sequence.
        When this event is set, the sequence manager will attempt to stop the sequence
        as soon as possible.
        Note that the sequence manager cannot interrupt a shot that is currently
        running, but will wait for it to finish.
        shot_retry_config: Specifies how to retry a shot if an error occurs.
        If an error occurs when the shot runner is running a shot, it will be caught
        by the sequence manager and the shot will be retried according to the
        configuration in this object.
        device_configurations: The device configurations to use to
        run the sequence.
        If None, the sequence manager will use the default device configurations.
    """

    def __init__(
        self,
        sequence: PureSequencePath,
        session_maker: ExperimentSessionMaker,
        interruption_event: threading.Event,
        shot_retry_config: Optional[ShotRetryConfig],
        global_parameters: Optional[ParameterNamespace],
        device_configurations: Optional[Mapping[DeviceName, DeviceConfiguration]],
        device_manager_extension: DeviceManagerExtensionProtocol,
    ) -> None:
        self._session_maker = session_maker
        self._sequence_path = sequence
        self._shot_retry_config = shot_retry_config or ShotRetryConfig()

        with self._session_maker() as session:
            if device_configurations is None:
                self.device_configurations = dict(session.default_device_configurations)
            else:
                self.device_configurations = dict(device_configurations)
            if global_parameters is None:
                self.sequence_parameters = session.get_global_parameters()
            else:
                self.sequence_parameters = copy.deepcopy(global_parameters)
            self.time_lanes = session.sequences.get_time_lanes(self._sequence_path)

        self._interruption_event = interruption_event

        self._device_manager_extension = device_manager_extension
        self._device_compilers: dict[DeviceName, DeviceCompiler] = {}

        self._watch_for_interruption_scope = anyio.CancelScope()

    @contextlib.asynccontextmanager
    async def run_sequence(self) -> AsyncGenerator[ShotScheduler, None]:
        """Run background tasks to compile and run shots for a given sequence.

        Returns:
            A asynchronous context manager that yields a shot scheduler object.

            When the context manager is entered, it will set the sequence to PREPARING
            while acquiring the necessary resources and the transition to RUNNING.

            The context manager will yield a shot scheduler object that can be used to
            push shots to the sequence execution queue.
            When a shot is done, its associated data will be stored in the associated
            sequence.

            One shot scheduling is over, the context manager will be exited.
            At this point is will finish the sequence and transition the sequence state
            to FINISHED when the sequence terminated normally, CRASHED if an error
            occurred or INTERRUPTED if the sequence was interrupted by the user.
        """

        self._prepare_sequence()
        try:
            shot_compiler = create_shot_compiler(
                SequenceContext(
                    device_configurations=self.device_configurations,
                    time_lanes=self.time_lanes,
                ),
                self._device_manager_extension,
            )
            async with (
                create_shot_runner(
                    shot_compiler, self._device_manager_extension
                ) as shot_runner,
                ShotManager(shot_runner, shot_compiler, self._shot_retry_config) as (
                    scheduler_cm,
                    data_stream_cm,
                ),
            ):
                self._set_sequence_state(State.RUNNING)
                async with (
                    anyio.create_task_group() as tg,
                    scheduler_cm as scheduler,
                ):
                    tg.start_soon(self._watch_for_interruption)
                    tg.start_soon(self._store_shots, data_stream_cm)
                    yield scheduler
        except* SequenceInterruptedException:
            self._set_sequence_state(State.INTERRUPTED)
            raise
        except* BaseException:
            self._set_sequence_state(State.CRASHED)
            raise
        else:
            self._set_sequence_state(State.FINISHED)

    def _prepare_sequence(self):
        with self._session_maker() as session:
            session.sequences.set_state(self._sequence_path, State.PREPARING)
            session.sequences.set_device_configurations(
                self._sequence_path, self.device_configurations
            )
            session.sequences.set_global_parameters(
                self._sequence_path, self.sequence_parameters
            )

    def _set_sequence_state(self, state: State):
        with self._session_maker() as session:
            session.sequences.set_state(self._sequence_path, state)

    async def _watch_for_interruption(self):
        with self._watch_for_interruption_scope:
            while True:
                if self._interruption_event.is_set():
                    raise SequenceInterruptedException(
                        f"Sequence '{self._sequence_path}' received an external "
                        f"interruption signal."
                    )
                await anyio.sleep(20e-3)

    async def _store_shots(
        self,
        data_stream_cm: contextlib.AbstractAsyncContextManager[AsyncIterable[ShotData]],
    ):
        async with data_stream_cm as shots_data:
            async for shot_data in shots_data:
                self._store_shot(shot_data)
        self._watch_for_interruption_scope.cancel()

    def _store_shot(self, shot_data: ShotData) -> None:
        params = {
            name: value for name, value in shot_data.variables.to_flat_dict().items()
        }
        with self._session_maker() as session:
            session.sequences.create_shot(
                self._sequence_path,
                shot_data.index,
                params,
                shot_data.data,
                shot_data.start_time,
                shot_data.end_time,
            )



import threading
from collections.abc import Mapping
from typing import Optional

from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtWidgets import QWidget, QTableView, QVBoxLayout

from analyza.loading.importers import ParametersImporter
from caqtus.session import ExperimentSession, get_standard_experiment_session
from caqtus.session.sequence import Shot
from caqtus.types import Parameter
from caqtus.utils import attrs, serialization
from ..single_shot_view import ShotView

ParameterName = str


@serialization.customize(
    _importer=serialization.override(rename="importer"),
)
@attrs.define(slots=False, init=False)
class ParametersView(ShotView):
    _importer: ParametersImporter = attrs.field(default=None)

    def __init__(
        self,
        *,
        importer: ParametersImporter,
        session: Optional[ExperimentSession] = None,
    ):
        super().__init__()

        if session is None:
            session = get_standard_experiment_session()

        self._importer = importer
        self._session = session
        self._lock = threading.Lock()

        self._setup_ui()

    def _setup_ui(self) -> None:
        self._model = ParamsModel()
        self._view = QTableView()
        self._view.setModel(self._model)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self._view)

    def set_shot(self, shot: Shot) -> None:
        with self._lock, self._session.activate():
            params = self._importer(shot, self._session)
        self._model.set_params(params)

    def update_view(self) -> None:
        pass


class ParamsModel(QAbstractTableModel):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)

        self._params: dict[ParameterName, Parameter] = {}

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._params)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 2

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return str(list(self._params.keys())[index.row()])
            elif index.column() == 1:
                return str(list(self._params.values())[index.row()])
        return None

    def set_params(self, params: Mapping[ParameterName, Parameter]) -> None:
        keys_values = zip(params.keys(), params.values())
        params = dict(sorted(keys_values, key=lambda x: x[0]))
        self.beginResetModel()
        self._params = dict(params)
        self.endResetModel()

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                if section == 0:
                    return "Name"
                elif section == 1:
                    return "Value"
        return super().headerData(section, orientation, role)



import anyio

from ._proxy import SequencerProxy
from .instructions import SequencerInstruction
from .trigger import SoftwareTrigger
from .._controller import DeviceController


class SequencerController(DeviceController):
    """Controls a sequencer during a shot."""

    async def run_shot(
        self,
        sequencer: SequencerProxy,
        /,
        sequence: SequencerInstruction,
        *args,
        **kwargs,
    ) -> None:
        await sequencer.update_parameters(sequence=sequence, *args, **kwargs)
        await self.start(sequencer)
        await self.wait_until_finished(sequencer)

    async def start(self, sequencer: SequencerProxy) -> None:
        trigger = await sequencer.get_trigger()
        if isinstance(trigger, SoftwareTrigger):
            await self.wait_all_devices_ready()
            await sequencer.start_sequence()
        else:
            await sequencer.start_sequence()
            await self.wait_all_devices_ready()

    async def wait_until_finished(self, sequencer: SequencerProxy) -> None:
        # We shield the task because we don't want the sequence to be stopped in the
        # middle with possibly dangerous values on the sequencer channels.
        with anyio.CancelScope(shield=True):
            while not await sequencer.has_sequence_finished():
                await self.sleep(0)



from typing import TypeAlias

import attrs


class LocalExperimentManagerConfiguration:
    pass


@attrs.define
class RemoteExperimentManagerConfiguration:
    address: str
    port: int
    authkey: str


ExperimentManagerConnection: TypeAlias = (
    LocalExperimentManagerConfiguration | RemoteExperimentManagerConfiguration
)



from typing import Optional

from PySide6.QtWidgets import QDialog, QWidget

from analyza.loading.importers import ParametersLoader
from .create_parameters_viewer_ui import Ui_ParametersDialog
from .params_viewer import ParametersView


def create_parameters_viewer(
    parent: Optional[QWidget],
) -> Optional[tuple[str, ParametersView]]:
    dialog = ViewerDialog(parent)

    if dialog.exec() == QDialog.DialogCode.Accepted:
        return (
            dialog.get_title(),
            ParametersView(
                importer=ParametersLoader(),
            ),
        )
    return None


class ViewerDialog(QDialog, Ui_ParametersDialog):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setupUi(self)

    def get_title(self) -> str:
        return self._window_title.text()



from __future__ import annotations

import datetime
from typing import Optional

import anyio
import anyio.to_thread
import attrs
import polars
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget
from caqtus.analysis.loading import DataImporter
from caqtus.session import (
    PureSequencePath,
    ExperimentSessionMaker,
    Shot,
)
from caqtus.session._return_or_raise import unwrap
from caqtus.session.path_hierarchy import PathNotFoundError
from caqtus.session.sequence_collection import PathIsNotSequenceError, PureShot
from caqtus.utils.itertools import batched

from .loader_ui import Ui_Loader


class DataLoader(QWidget, Ui_Loader):
    def __init__(
        self,
        shot_loader: DataImporter,
        session_maker: ExperimentSessionMaker,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self.setupUi(self)

        self.watchlist: dict[PureSequencePath, SequenceLoadingInfo] = {}
        self.session_maker = session_maker
        self.process_chunk_size = 10
        self.shot_loader = shot_loader
        self.clear_button.clicked.connect(self.clear_watchlist)

    def add_sequence_to_watchlist(self, sequence_path: PureSequencePath):
        if sequence_path not in self.watchlist:
            with self.session_maker() as session:
                stats = unwrap(session.sequences.get_stats(sequence_path))
                start_time = stats.start_time
                number_completed_shots = stats.number_completed_shots
            self.watchlist[sequence_path] = SequenceLoadingInfo(
                start_time=start_time,
                number_completed_shots=number_completed_shots,
                processed_shots=set(),
                dataframe=empty_dataframe(),
            )
            self.sequence_list.addItem(str(sequence_path))

    def clear_watchlist(self):
        self.watchlist = {}
        self.sequence_list.clear()

    def get_sequences_data(self) -> dict[PureSequencePath, polars.DataFrame]:
        return {path: info.dataframe for path, info in self.watchlist.items()}

    def remove_sequence_from_watchlist(self, sequence_path: PureSequencePath) -> None:
        """Remove a sequence from the watchlist.

        This method has no effect if the sequence is not in the watchlist.
        """

        if sequence_path in self.watchlist:
            self.watchlist.pop(sequence_path)
            # Need to use reversed to avoid index shifting when removing items
            for item in reversed(
                self.sequence_list.findItems(
                    str(sequence_path), Qt.MatchFlag.MatchExactly
                )
            ):
                self.sequence_list.takeItem(self.sequence_list.row(item))

    def update_progress_bar(self) -> None:
        """Update the progress bar to reflect the number of processed shots."""

        total_shots = sum(
            info.number_completed_shots for info in self.watchlist.values()
        )
        processed_shots = sum(
            len(info.processed_shots) for info in self.watchlist.values()
        )
        self.progress_bar.setMaximum(total_shots)
        self.progress_bar.setValue(processed_shots)

    async def process(self):
        while True:
            await self.single_process()
            await anyio.sleep(1e-3)

    async def single_process(self):
        # Can't update over the dict watchlist, because it might be updated during the processing
        for sequence_path in list(self.watchlist):
            await self.process_new_shots(sequence_path)

    async def process_new_shots(self, sequence: PureSequencePath) -> None:
        async with self.session_maker.async_session() as session:
            # Check if the sequence has been reset by comparing the start time
            # of the sequence in the watchlist with the start time of the sequence in the session.
            # If the start time is different, the sequence has been reset, and we clear the processed shots.
            stats_result = await session.sequences.get_stats(sequence)
            try:
                stats = unwrap(stats_result)
            except (PathNotFoundError, PathIsNotSequenceError):
                self.remove_sequence_from_watchlist(sequence)
                return
            try:
                loading_info = self.watchlist[sequence]
            except KeyError:
                return
            if stats.start_time != loading_info.start_time:
                self.watchlist[sequence] = SequenceLoadingInfo(
                    start_time=stats.start_time,
                    number_completed_shots=stats.number_completed_shots,
                    processed_shots=set(),
                    dataframe=empty_dataframe(),
                )
                return
            result = await session.sequences.get_shots(sequence)
            try:
                shots: list[PureShot] = unwrap(result)
            except (PathNotFoundError, PathIsNotSequenceError):
                self.remove_sequence_from_watchlist(sequence)
                return

        try:
            processed_shots = self.watchlist[sequence].processed_shots
        except KeyError:
            return
        new_shots = sorted(
            (shot for shot in shots if shot.index not in processed_shots),
            key=lambda s: s.index,
        )

        for shot_group in batched(new_shots, self.process_chunk_size):
            with self.session_maker() as session:
                for shot in shot_group:
                    await self.process_shot(Shot.bound(shot, session))
                # self.update_progress_bar()

    async def process_shot(self, shot: Shot) -> None:
        new_data = await anyio.to_thread.run_sync(self.shot_loader, shot)
        try:
            processing_info = self.watchlist[shot.sequence.path]
        except KeyError:
            return
        total_data = processing_info.dataframe
        concatenated = polars.concat([total_data, new_data])
        processing_info.dataframe = concatenated
        processing_info.processed_shots.add(shot.index)


@attrs.define
class SequenceLoadingInfo:
    start_time: Optional[datetime.datetime]
    number_completed_shots: int
    processed_shots: set[int]
    dataframe: polars.DataFrame


def empty_dataframe() -> polars.DataFrame:
    return polars.DataFrame()



from ._compiler import SequencerCompiler
from ..channel_commands._channel_sources._trigger_compiler import (
    TriggerableDeviceCompiler,
)

__all__ = ["SequencerCompiler", "TriggerableDeviceCompiler"]



import contextlib
import functools
from collections.abc import Callable, Coroutine, Iterator
from typing import (
    Self,
    ParamSpec,
    TypeVar,
    Generic,
    LiteralString,
    Any,
    final,
    AsyncIterator,
)

from .rpc import RPCClient, Proxy, RemoteCallError
from ..runtime import Device
from ...utils.contextlib import aclose_on_error

T = TypeVar("T")
P = ParamSpec("P")

DeviceType = TypeVar("DeviceType", bound=Device)


def unwrap_remote_error_decorator(fun: Callable[P, Coroutine[Any, Any, T]]):
    """Decorator that unwraps RemoteError and raises the original exception.

    It pretends that the original exception occurred on the client side.

    It is useful to help catch device specific exceptions for example timeouts, since
    then we can directly catch TimeoutError and not RemoteError.
    """

    @functools.wraps(fun)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        with unwrap_remote_error_cm():
            return await fun(*args, **kwargs)

    return wrapper


@contextlib.contextmanager
def unwrap_remote_error_cm():
    try:
        yield
    except RemoteCallError as remote:
        if original := remote.__cause__:
            # Here we need to break the circle of exceptions.
            # Indeed, original.__context__ is now remote, and remote.__cause__ is
            # original.
            # If not handled, this would cause recursion issues when anything tries
            # to handle the exception down the line.
            remote.__cause__ = None
            raise original
        else:
            raise


async def unwrap_remote_error_iterator(iterator: AsyncIterator[T]) -> AsyncIterator[T]:
    stop = object()
    while True:
        with unwrap_remote_error_cm():
            next_value = await anext(iterator, stop)
        if next_value is stop:
            break
        yield next_value


class DeviceProxy(Generic[DeviceType]):
    """Proxy to a remote device.

    This class is used on the client side to interact with a device running on a remote
    server.
    It provides asynchronous methods to get attributes and call methods remotely
    without blocking the client.
    """

    @final
    def __init__(
        self,
        rpc_client: RPCClient,
        device_type: Callable[P, DeviceType],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self._rpc_client = rpc_client
        self._device_type = device_type
        self._args = args
        self._kwargs = kwargs
        self._device_proxy: Proxy[DeviceType]

        self._async_exit_stack = contextlib.AsyncExitStack()

    async def __aenter__(self) -> Self:
        async with aclose_on_error(self._async_exit_stack):
            self._device_proxy = await self._async_exit_stack.enter_async_context(
                self._rpc_client.call_proxy_result(
                    self._device_type, *self._args, **self._kwargs
                )
            )
            with unwrap_remote_error_cm():
                await self._async_exit_stack.enter_async_context(
                    self.async_context_manager(self._device_proxy)
                )
        return self

    @unwrap_remote_error_decorator
    async def get_attribute(self, attribute_name: LiteralString) -> Any:
        return await self._rpc_client.get_attribute(self._device_proxy, attribute_name)

    @unwrap_remote_error_decorator
    async def call_method(
        self,
        method_name: LiteralString,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return await self._rpc_client.call_method(
            self._device_proxy, method_name, *args, **kwargs
        )

    def call_method_proxy_result(
        self,
        method_name: LiteralString,
        *args: Any,
        **kwargs: Any,
    ) -> contextlib.AbstractAsyncContextManager[Proxy]:
        return self._rpc_client.call_method_proxy_result(
            self._device_proxy, method_name, *args, **kwargs
        )

    def async_context_manager(
        self, proxy: Proxy[contextlib.AbstractContextManager[T]]
    ) -> contextlib.AbstractAsyncContextManager[Proxy[T]]:
        return self._rpc_client.async_context_manager(proxy)

    def async_iterator(self, proxy: Proxy[Iterator[T]]) -> AsyncIterator[T]:
        return unwrap_remote_error_iterator(self._rpc_client.async_iterator(proxy))

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self._async_exit_stack.__aexit__(exc_type, exc_value, traceback)



"""Contains the main GUI from which user can edit and launch sequences."""

from . import device_configuration_editors, _parameter_tables_editor

__all__ = ["device_configuration_editors", "parameter_tables_editor"]



import threading
from typing import Optional

from PySide6.QtWidgets import QVBoxLayout
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure

from analyza.loading.importers import AtomImporter2D
from caqtus.session import ExperimentSession, get_standard_experiment_session
from caqtus.session.sequence import Shot
from caqtus.utils import attrs, serialization
from ..single_shot_view import ShotView


@serialization.customize(
    _importer=serialization.override(rename="importer"),
    _axes_equal=serialization.override(rename="axes_equal"),
)
@attrs.define(slots=False, init=False)
class AtomsView(ShotView):
    _importer: AtomImporter2D = attrs.field(default=None)
    _axes_equal: bool = attrs.field(default=True)

    def __init__(
        self,
        *,
        importer: AtomImporter2D,
        axes_equal: bool = True,
        session: Optional[ExperimentSession] = None,
    ):
        super().__init__()

        self.__attrs_init__(importer=importer, axes_equal=axes_equal)

        if session is None:
            session = get_standard_experiment_session()

        self._session = session

        self._lock = threading.Lock()

        self._setup_ui()

    def _setup_ui(self) -> None:
        self._figure = Figure()
        self._axes = self._figure.add_subplot()
        if self._axes_equal:
            self._axes.set_aspect("equal")
        self._canvas = FigureCanvasQTAgg(self._figure)

        self.setLayout(QVBoxLayout())
        navigation_toolbar = NavigationToolbar2QT(self._canvas, self)
        self.layout().addWidget(navigation_toolbar)
        self.layout().addWidget(self._canvas)

    def set_shot(self, shot: Shot) -> None:
        with self._lock, self._session.activate():
            try:
                atoms = self._importer(shot, self._session)
            except Exception as e:
                self._set_exception(e)
            else:
                self._paint_atoms(atoms)

    def update_view(self) -> None:
        self._canvas.draw()

    def _paint_atoms(self, atoms: dict[tuple[float, float], bool]) -> None:
        self._axes.clear()
        for (x, y), atom in atoms.items():
            if atom:
                self._axes.plot(x, y, "o", color="black")
            else:
                self._axes.plot(x, y, "o", color="black", alpha=0.1)

    def _set_exception(self, error: Exception):
        self._axes.clear()
        self._axes.text(
            0.5,
            0.5,
            f"{error!r}",
            horizontalalignment="center",
            verticalalignment="center",
            color="red",
        )
        self._canvas.draw()



from typing import Optional

import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from ._instructions import SequencerInstruction
from ._to_time_array import convert_to_change_arrays


def plot_instruction(
    instruction: SequencerInstruction, ax: Optional[Axes] = None
) -> Axes:
    """Plot the instruction on the given axis."""

    if ax is None:
        fig, ax = plt.subplots()

    change_times, change_values = convert_to_change_arrays(instruction)

    ax.step(change_times, change_values, where="post")
    ax.set_xlabel("Time [ticks]")
    ax.set_ylabel("Value")
    return ax



import attrs as attrs

from ._add_exc_note import add_exc_note
from ._log_duration import log_duration
from ._log_exception import log_exception
from .duration_timer import DurationTimer, DurationTimerLog

__all__ = [
    "log_exception",
    "log_duration",
    "DurationTimer",
    "DurationTimerLog",
    "attrs",
    "add_exc_note",
]



from .units import (
    extract_unit,
    extract_units,
    add_unit,
    convert_to_unit,
    magnitude_in_unit,
    with_columns_expressed_in_units,
    with_units_added_to_columns,
    MAGNITUDE_FIELD,
    UNITS_FIELD,
    is_quantity_dtype,
)

__all__ = [
    "extract_unit",
    "extract_units",
    "add_unit",
    "convert_to_unit",
    "magnitude_in_unit",
    "with_columns_expressed_in_units",
    "with_units_added_to_columns",
    "MAGNITUDE_FIELD",
    "UNITS_FIELD",
    "is_quantity_dtype",
]



from .visualizer_creator import DataView, ViewCreator
from .visualizer_creators_selector import VisualizerCreatorSelector
from .visualizers import ErrorBarViewCreator, DataFrameViewCreator

__all__ = [
    "DataView",
    "ViewCreator",
    "VisualizerCreatorSelector",
    "ErrorBarViewCreator",
    "DataFrameViewCreator",
]



from __future__ import annotations

from typing import Optional, Mapping, Any

import attrs

import caqtus.formatter as fmt
from caqtus.shot_compilation import ShotContext
from caqtus.types.recoverable_exceptions import InvalidValueError, InvalidTypeError
from caqtus.types.timelane import DigitalTimeLane, AnalogTimeLane
from caqtus.types.variable_name import DottedVariableName
from ._compile_digital_lane import compile_digital_lane
from .compile_analog_lane import compile_analog_lane
from ..channel_output import ChannelOutput, DimensionedSeries
from ..._time_step import TimeStep
from ...instructions import Pattern


@attrs.define
class LaneValues(ChannelOutput):
    """Indicates that the output should be the values taken by a given lane.

    Attributes:
        lane: The name of the lane from which to take the values.
        default: The default value to take if the lane is absent from the shot
            time lanes.
    """

    lane: str = attrs.field(
        converter=str,
        on_setattr=attrs.setters.convert,
    )
    default: Optional[ChannelOutput] = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.instance_of(ChannelOutput)
        ),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self) -> str:
        if self.default is not None:
            return f"{self.lane} | {self.default}"
        return self.lane

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ):
        """Evaluate the output of a channel as the values of a lane.

        This function will look in the shot time lanes to find the lane referenced by
        the output and evaluate the values of this lane.
        If the lane cannot be found, and the output has a default value, this default
        value will be used.
        If the lane cannot be found and there is no default value, a ValueError will be
        raised.
        """

        lane_name = self.lane
        try:
            lane = shot_context.get_lane(lane_name)
        except KeyError:
            if self.default is not None:
                return self.default.evaluate(
                    required_time_step,
                    prepend,
                    append,
                    shot_context,
                )
            else:
                raise InvalidValueError(
                    f"Could not find {fmt.lane(lane_name)}"
                ) from None
        if isinstance(lane, DigitalTimeLane):
            lane_values = compile_digital_lane(lane, required_time_step, shot_context)
            result = DimensionedSeries(lane_values, units=None)
        elif isinstance(lane, AnalogTimeLane):
            result = compile_analog_lane(
                lane,
                shot_context.get_variables(),
                shot_context.get_step_start_times(),
                required_time_step,
            )
        else:
            raise InvalidTypeError(
                f"Don't know how to compile lane {fmt.lane(lane_name)} with type "
                f"{fmt.type_(type(lane))}"
            )

        prepend_value = result.values[0]
        prepend_pattern = prepend * Pattern([prepend_value])
        append_value = result.values[-1]
        append_pattern = append * Pattern([append_value])
        return DimensionedSeries(
            values=prepend_pattern + result.values + append_pattern,
            units=result.units,
        )

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        return 0, 0



import abc
from collections.abc import Iterator
from contextlib import AbstractContextManager
from typing import ClassVar

from attrs import define, field
from attrs.setters import frozen, convert
from attrs.validators import instance_of
from caqtus.device import Device
from caqtus.types.image import Image

from ._configuration import RectangularROI


class CameraTimeoutError(TimeoutError):
    """Raised when the camera did not acquire an image after a given timeout."""

    pass


@define(slots=False)
class Camera(Device, abc.ABC):
    """Define the interface for a camera instrument.

    This is an abstract class that must be subclassed to implement a specific camera.
    Subclasses must implement the :meth:`update_parameters` and :meth:`acquire` methods.

    Fields:
        roi: A rectangular subset of the sensor that defines the region of interest.
            Images returned by the camera must be of the same size as the region of
            interest.
            Some camera models allow to define the region of interest before even
            fetching the image from the instrument.
            This is the recommended way to crop the region of interest, as it will
            reduce the amount of data transferred from the camera to the computer.
            If the camera does not support defining the region of interest before
            fetching the image, the region of interest should be cropped after the image
            is acquired and before it is returned to the user.
            This attribute can only be set with the :meth:`__init__` method.
        timeout: The maximum time in seconds that the camera must wait for an external
            trigger signal before raising a :class:`CameraTimeoutError`.
        external_trigger: A boolean that indicates if the camera is waiting for an
            external trigger signal to acquire an image.
            If this is set to True, the camera will wait for the trigger signal before
            acquiring an image.
            If this is set to False, the camera will acquire an image as soon as
            possible after the acquisition is started.
            This attribute can only be set with the :meth:`__init__` method.
        sensor_width: A class attribute that defines the width of the sensor in pixels.
            This attribute must be set in the subclass implementation.
        sensor_height: A class attribute that defines the height of the sensor in
            pixels.
            This attribute must be set in the subclass implementation.
    """

    sensor_width: ClassVar[int]
    sensor_height: ClassVar[int]

    roi: RectangularROI = field(
        validator=instance_of(RectangularROI), on_setattr=frozen
    )
    timeout: float = field(converter=float, on_setattr=convert)
    external_trigger: bool = field(validator=instance_of(bool), on_setattr=frozen)

    @roi.validator  # type: ignore
    def _validate_roi(self, _, value: RectangularROI):
        if value.original_image_size != (self.sensor_width, self.sensor_height):
            raise ValueError(
                f"The original image size of the ROI {value.original_image_size} "
                f"does not match the sensor size "
                f"{self.sensor_width}x{self.sensor_height}"
            )

    @abc.abstractmethod
    def update_parameters(self, timeout: float, *args, **kwargs) -> None:
        """Update the camera parameters between acquisitions.

        It is undefined what should happen if this method is called while the camera is
        acquiring images.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def acquire(
        self, exposures: list[float]
    ) -> AbstractContextManager[Iterator[Image]]:
        """Acquire images with the given exposure times.

        Returns:
            A context manager that yields an iterator of images.
            When the context manager is entered, the camera starts the acquisition of
            the number of images specified by the length of the exposures list.

            Iterating over the iterator returned by the context manager will yield the
            images as they are acquired by the camera.
            It is recommended to yield the images as they are acquired, and not to wait
            for all the images to be acquired before returning them.
            This allows to do some processing on the images and to react to them while
            the camera is still acquiring the next ones.

            When the context manager exits, the camera stops the acquisition.

            Note that not all images might have been acquired when the context manager
            exits, if not all images have been consumed by the user due to an exception
            or a break in the with statement.

        Raises:
            CameraTimeoutError: In the iterator if the camera could not acquire an
                image after the timeout specified by the method
                :meth:`update_parameters`.

        Example:

            This demonstrates how to use this method to take images:

            .. code-block:: python

                with camera.acquire(exposures=[0.1, 0.5, 1.0]) as images:
                    for image in images:
                        print(image)


        """

        raise NotImplementedError

    def take_picture(self, exposure: float) -> Image:
        """A convenience method to acquire a single image."""

        with self.acquire([exposure]) as images:
            return next(iter(images))



import contextlib

import anyio


@contextlib.contextmanager
def renamed_exception_group(message: str):
    try:
        yield
    except ExceptionGroup as e:
        raise ExceptionGroup(message, e.exceptions) from None


@contextlib.asynccontextmanager
async def task_group_with_error_message(message: str):
    with renamed_exception_group(message):
        async with anyio.create_task_group() as tg:
            yield tg


from __future__ import annotations

import datetime
import typing
from collections.abc import Mapping, Set

import attrs

from caqtus.types.data import DataLabel, Data
from caqtus.types.parameter import Parameter
from caqtus.types.variable_name import DottedVariableName
from ..path import PureSequencePath
from ..sequence_collection import PureShot

# We don't do these imports at runtime because it would create a circular import.
if typing.TYPE_CHECKING:
    from .sequence import Sequence
    from ..experiment_session import ExperimentSession


@attrs.frozen(eq=False, order=False)
class Shot:
    """Gives access to the data of a shot."""

    sequence_path: PureSequencePath
    index: int
    _session: ExperimentSession

    # Since the session implements repeatable reads, we can cache the data that we have
    # already fetched and know that they are up-to-date.
    # Warning: This only works if the current session does not update the shot, which
    # should not happen on the user side.
    _data_cache: dict[DataLabel, Data] = attrs.field(factory=dict)

    @classmethod
    def bound(cls, shot: PureShot, session: ExperimentSession) -> typing.Self:
        return cls(shot.sequence_path, shot.index, session)

    @property
    def sequence(self) -> "Sequence":
        """The sequence to which this shot belongs."""

        from .sequence import Sequence

        return Sequence(self.sequence_path, self._session)

    def get_parameters(self) -> Mapping[DottedVariableName, Parameter]:
        """Return the parameters used to run this shot."""

        return self._session.sequences.get_shot_parameters(
            self.sequence_path, self.index
        )

    def get_data(self) -> Mapping[DataLabel, Data]:
        """Return the data of this shot.

        This will return all data that was acquired during the shot.
        If you want to get only a subset of the data, use :meth:`get_data_by_label`
        which will avoid querying unnecessary data.
        """

        data = self._session.sequences.get_all_shot_data(self.sequence_path, self.index)
        self._data_cache.update(data)
        return data

    def get_data_by_label(self, label: DataLabel) -> Data:
        """Return the data of this shot with the given label."""

        if label in self._data_cache:
            return self._data_cache[label]
        else:
            data = self._session.sequences.get_shot_data_by_label(
                self.sequence_path, self.index, label
            )
            self._data_cache[label] = data
            return data

    def get_data_by_labels(self, labels: Set[DataLabel]) -> Mapping[DataLabel, Data]:
        """Return the data of this shot with the given labels."""

        cached = set(self._data_cache.keys())
        to_fetch = labels - cached

        fetched = self._session.sequences.get_shot_data_by_labels(
            self.sequence_path, self.index, to_fetch
        )
        self._data_cache.update(fetched)

        return {label: self._data_cache[label] for label in labels}

    def get_start_time(self) -> datetime.datetime:
        """Return the time at which this shot started running."""

        return self._session.sequences.get_shot_start_time(
            self.sequence_path, self.index
        )

    def get_end_time(self) -> datetime.datetime:
        """Return the time at which this shot finished running."""

        return self._session.sequences.get_shot_end_time(self.sequence_path, self.index)



from ._trigger import (
    Trigger,
    SoftwareTrigger,
    ExternalTriggerStart,
    ExternalClock,
    ExternalClockOnChange,
    TriggerEdge,
    is_trigger,
)

__all__ = [
    "Trigger",
    "SoftwareTrigger",
    "ExternalClock",
    "ExternalTriggerStart",
    "ExternalClockOnChange",
    "TriggerEdge",
    "is_trigger",
]



from copy import deepcopy
from typing import Generic, TypeVar, Self

from caqtus.shot_compilation import VariableNamespace
from caqtus.types.variable_name import DottedVariableName

T = TypeVar("T")


class StepContext(Generic[T]):
    """Immutable context that contains the variables of a given step."""

    def __init__(self) -> None:
        self._variables = VariableNamespace[T]()

    def clone(self) -> Self:
        return deepcopy(self)

    def update_variable(self, name: DottedVariableName, value: T) -> Self:
        clone = self.clone()
        clone._variables.update({name: value})
        return clone

    @property
    def variables(self):
        return deepcopy(self._variables)



from typing import Mapping

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QWidget, QVBoxLayout

from caqtus.utils.itertools import first
from .visualizer_creator import DataView, ViewCreator
from .visualizer_creators_selector_ui import Ui_VisualizerCreatorSelector


class VisualizerCreatorSelector(QWidget, Ui_VisualizerCreatorSelector):
    visualizer_selected = pyqtSignal(DataView)

    def __init__(
        self, visualizer_creators: Mapping[str, ViewCreator], *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)

        self._visualizer_creators = dict(visualizer_creators)
        self._current_visualizer_creator: ViewCreator = first(
            visualizer_creators.values()
        )
        self._setup_ui()

    def _setup_ui(self) -> None:
        self.setupUi(self)
        for visualiser_creator in self._visualizer_creators:
            self._visualizer_combo_box.addItem(visualiser_creator)
        self._apply_button.clicked.connect(self._on_apply_button_clicked)
        self._settings_group.setHidden(True)
        self._visualizer_combo_box.currentTextChanged.connect(
            self._on_visualizer_creator_selected
        )
        self._settings_group.setLayout(QVBoxLayout())

    def _on_visualizer_creator_selected(self, visualizer_creator_name: str) -> None:
        self._current_visualizer_creator = self._visualizer_creators[
            visualizer_creator_name
        ]
        layout = self._settings_group.layout()
        while layout.count() > 0:
            layout.itemAt(0).widget().setParent(None)
        if isinstance(self._current_visualizer_creator, QWidget):
            self._settings_group.layout().addWidget(self._current_visualizer_creator)
            self._settings_group.setHidden(False)
        else:
            self._settings_group.setHidden(True)

    def _on_apply_button_clicked(self) -> None:
        self.visualizer_selected.emit(self._current_visualizer_creator.create_view())  # type: ignore



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'editor.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QWidget,
)


class Ui_CameraConfigurationEditor(object):
    def setupUi(self, CameraConfigurationEditor):
        if not CameraConfigurationEditor.objectName():
            CameraConfigurationEditor.setObjectName("CameraConfigurationEditor")
        CameraConfigurationEditor.resize(424, 549)
        self.horizontalLayout = QHBoxLayout(CameraConfigurationEditor)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.form = QFormLayout()
        self.form.setObjectName("form")
        self.label_2 = QLabel(CameraConfigurationEditor)
        self.label_2.setObjectName("label_2")

        self.form.setWidget(0, QFormLayout.LabelRole, self.label_2)

        self.formLayout_3 = QFormLayout()
        self.formLayout_3.setObjectName("formLayout_3")
        self.formLayout_3.setVerticalSpacing(6)
        self.formLayout_3.setContentsMargins(-1, 5, -1, -1)
        self.xLabel = QLabel(CameraConfigurationEditor)
        self.xLabel.setObjectName("xLabel")

        self.formLayout_3.setWidget(0, QFormLayout.LabelRole, self.xLabel)

        self._left_spinbox = QSpinBox(CameraConfigurationEditor)
        self._left_spinbox.setObjectName("_left_spinbox")

        self.formLayout_3.setWidget(0, QFormLayout.FieldRole, self._left_spinbox)

        self.widthLabel = QLabel(CameraConfigurationEditor)
        self.widthLabel.setObjectName("widthLabel")

        self.formLayout_3.setWidget(1, QFormLayout.LabelRole, self.widthLabel)

        self._right_spinbox = QSpinBox(CameraConfigurationEditor)
        self._right_spinbox.setObjectName("_right_spinbox")

        self.formLayout_3.setWidget(1, QFormLayout.FieldRole, self._right_spinbox)

        self.yLabel = QLabel(CameraConfigurationEditor)
        self.yLabel.setObjectName("yLabel")

        self.formLayout_3.setWidget(2, QFormLayout.LabelRole, self.yLabel)

        self._bottom_spinbox = QSpinBox(CameraConfigurationEditor)
        self._bottom_spinbox.setObjectName("_bottom_spinbox")

        self.formLayout_3.setWidget(2, QFormLayout.FieldRole, self._bottom_spinbox)

        self.heightLabel = QLabel(CameraConfigurationEditor)
        self.heightLabel.setObjectName("heightLabel")

        self.formLayout_3.setWidget(3, QFormLayout.LabelRole, self.heightLabel)

        self._top_spinbox = QSpinBox(CameraConfigurationEditor)
        self._top_spinbox.setObjectName("_top_spinbox")

        self.formLayout_3.setWidget(3, QFormLayout.FieldRole, self._top_spinbox)

        self.form.setLayout(0, QFormLayout.FieldRole, self.formLayout_3)

        self.horizontalLayout.addLayout(self.form)

        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.horizontalLayout.setStretch(1, 1)

        self.retranslateUi(CameraConfigurationEditor)

        QMetaObject.connectSlotsByName(CameraConfigurationEditor)

    # setupUi

    def retranslateUi(self, CameraConfigurationEditor):
        CameraConfigurationEditor.setWindowTitle(
            QCoreApplication.translate("CameraConfigurationEditor", "Form", None)
        )
        # if QT_CONFIG(tooltip)
        self.label_2.setToolTip(
            QCoreApplication.translate(
                "CameraConfigurationEditor",
                "<html><head/><body><p>Rectangular region of interest to extract from the full sensor picture.</p></body></html>",
                None,
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.label_2.setText(
            QCoreApplication.translate("CameraConfigurationEditor", "ROI", None)
        )
        self.xLabel.setText(
            QCoreApplication.translate("CameraConfigurationEditor", "Left", None)
        )
        self.widthLabel.setText(
            QCoreApplication.translate("CameraConfigurationEditor", "Right", None)
        )
        self.yLabel.setText(
            QCoreApplication.translate("CameraConfigurationEditor", "Bottom", None)
        )
        self.heightLabel.setText(
            QCoreApplication.translate("CameraConfigurationEditor", "Top", None)
        )

    # retranslateUi



import decimal
from abc import ABC, abstractmethod
from typing import ClassVar

import attrs

from caqtus.device.runtime import Device
from .._time_step import TimeStep
from ..instructions import SequencerInstruction
from ..trigger import Trigger, is_trigger


@attrs.define(slots=False)
class Sequencer(Device, ABC):
    """Base class for all sequencers.

    Fields:
        time_step: The time step of the sequencer in nanoseconds.
            This value cannot be changed after the sequencer has been created.
        trigger: Indicates how the sequence is started and how it is clocked.
            This value cannot be changed after the sequencer has been created.
    """

    channel_number: ClassVar[int]

    time_step: TimeStep = attrs.field(
        on_setattr=attrs.setters.frozen,
        converter=decimal.Decimal,
        validator=attrs.validators.gt(decimal.Decimal(0)),
    )
    trigger: Trigger = attrs.field(
        on_setattr=attrs.setters.frozen,
    )

    _sequence_programmed: bool = attrs.field(default=False, init=False)
    _sequence_started: bool = attrs.field(default=False, init=False)

    @trigger.validator  # type: ignore
    def _validate_trigger(self, _, value):
        if not is_trigger(value):
            raise ValueError(f"Invalid trigger {value}")

    @abstractmethod
    def update_parameters(
        self, sequence: SequencerInstruction, *args, **kwargs
    ) -> None:
        """Update the parameters of the sequencer.

        Args:
            sequence: The sequence to be programmed into the sequencer.
        """

        if sequence.width != self.channel_number:
            raise ValueError(
                f"Invalid number of channels, expected {self.channel_number} but got"
                f" {sequence.width}."
            )

    def _set_sequence_programmed(self) -> None:
        """To call after successful update_parameters."""

        self._sequence_started = False
        self._sequence_programmed = True

    @abstractmethod
    def start_sequence(self) -> None:
        """Start the sequence.

        To be subclassed by the specific sequencer implementation.
        The base class implementation checks if the sequence has been programmed and
        sets _sequence_started to True.

        Raises:
            SequenceNotConfiguredError: If the sequence has not been configured yet.
        """

        if not self._sequence_programmed:
            raise SequenceNotConfiguredError("The sequence has not been set yet.")

        self._sequence_started = True
        self._sequence_programmed = False

    @abstractmethod
    def has_sequence_finished(self) -> bool:
        """Check if the sequence has finished.

        Returns:
            True if the sequence has finished, False if it is still running.
        Raises:
            SequenceNotStartedError: If start_sequence has not been called yet.
        """

        if not self._sequence_started:
            raise SequenceNotStartedError("The sequence has not been started yet.")
        return True

    def wait_sequence_finished(self) -> None:
        while not self.has_sequence_finished():
            pass

    def get_trigger(self) -> Trigger:
        return self.trigger


class SequencerProgrammingError(RuntimeError):
    pass


class SequenceNotStartedError(SequencerProgrammingError):
    """Raised when the sequence as been configured, but has not been started yet."""

    pass


class SequenceNotConfiguredError(SequencerProgrammingError):
    """Raised when the sequence has not been configured yet."""

    pass



from ._configuration import (
    DeviceConfiguration,
    get_configurations_by_type,
    DeviceServerName,
    DeviceConfigType,
)
from ._parameter import DeviceParameter

__all__ = [
    "DeviceConfiguration",
    "get_configurations_by_type",
    "DeviceParameter",
    "DeviceServerName",
    "DeviceConfigType",
]



import logging
from functools import wraps
from typing import ParamSpec, TypeVar, Callable

_P = ParamSpec("_P")
_T = TypeVar("_T")


def log_exception(
    logger: logging.Logger,
    exec_info: bool = True,
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    """Decorator to log exceptions raised by a function.

    Args:
        logger: The logger to use to record the exception.
        exec_info: If True, the exception information will be added to the log record.
    """

    def decorator(func: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            try:
                return func(*args, **kwargs)
            except Exception:
                logger.exception(
                    f"An error occurred when executing {func.__name__}.",
                    exc_info=exec_info,
                )
                raise

        return wrapper

    return decorator



from PySide6.QtWidgets import QApplication

from caqtus.gui.condetrol.device_configuration_editors.sequencer_configuration_editor.channel_output_editor._analog_mapping_node import (
    CalibratedAnalogMappingWidget,
)

if __name__ == "__main__":
    app = QApplication([])

    widget = CalibratedAnalogMappingWidget()
    widget.set_input_units("dB")
    widget.set_data_points(
        [
            (-49.0, 0.5),
            (-36.0, 0.55),
            (-24.0, 0.6),
            (-13.0, 0.65),
            (-6.6, 0.7),
            (-2, 0.8),
            (-1, 0.9),
            (0, 1),
        ]
    )

    widget.show()

    app.exec()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'add_image_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Ui_ImageDialog(object):
    def setupUi(self, ImageDialog):
        if not ImageDialog.objectName():
            ImageDialog.setObjectName("ImageDialog")
        ImageDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ImageDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.titleLabel = QLabel(ImageDialog)
        self.titleLabel.setObjectName("titleLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.titleLabel)

        self._window_title = QLineEdit(ImageDialog)
        self._window_title.setObjectName("_window_title")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self._window_title)

        self.cameraLabel = QLabel(ImageDialog)
        self.cameraLabel.setObjectName("cameraLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.cameraLabel)

        self._camera_line_edit = QLineEdit(ImageDialog)
        self._camera_line_edit.setObjectName("_camera_line_edit")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self._camera_line_edit)

        self.pictureLabel = QLabel(ImageDialog)
        self.pictureLabel.setObjectName("pictureLabel")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.pictureLabel)

        self._picture_line_edit = QLineEdit(ImageDialog)
        self._picture_line_edit.setObjectName("_picture_line_edit")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self._picture_line_edit)

        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(ImageDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ImageDialog)
        self.buttonBox.accepted.connect(ImageDialog.accept)
        self.buttonBox.rejected.connect(ImageDialog.reject)

        QMetaObject.connectSlotsByName(ImageDialog)

    # setupUi

    def retranslateUi(self, ImageDialog):
        ImageDialog.setWindowTitle(
            QCoreApplication.translate("ImageDialog", "Dialog", None)
        )
        self.titleLabel.setText(
            QCoreApplication.translate("ImageDialog", "Title", None)
        )
        self.cameraLabel.setText(
            QCoreApplication.translate("ImageDialog", "Camera", None)
        )
        self.pictureLabel.setText(
            QCoreApplication.translate("ImageDialog", "Picture", None)
        )

    # retranslateUi



from __future__ import annotations

from collections.abc import Callable
from typing import Generic

import attrs

from caqtus.device import DeviceConfiguration, Device, DeviceController

# noinspection PyPep8Naming
from caqtus.device.configuration import DeviceConfigType as C
from caqtus.device.remote import DeviceProxy
from caqtus.gui.condetrol.device_configuration_editors import DeviceConfigurationEditor
from caqtus.shot_compilation import DeviceCompiler
from caqtus.utils.serialization import JSON


@attrs.frozen
class DeviceExtension(Generic[C]):
    """Extension to define how to implement a device in the application.

    This class is generic in the :class:`DeviceConfiguration` type.

    Attributes:
        label: A human-readable label for the type of device represented by this
            extension.

            This label describes the type of device and will be displayed to the user
            when they are selecting a device to add to the experiment.

            No two device extensions should have the same label.

        device_type: A callable that returns a new device instance.

            This function will be called when a device associated with this extension
            needs to be created.
            The arguments passed are obtained by calling the method
            :meth:`caqtus.device.compiler.DeviceCompiler.compile_initialization_parameters`
            of the compiler associated with the device.

            This function needs to be pickleable as it will be sent to a remote process.

        configuration_type: The type of configuration used to store the settings of
            the device.

            The name of the type is used to identify the configuration type.
            It is thus not recommended to change the name of the type after it has been
            used in the application, as the new name will not be recognized as the same
            type.

        configuration_factory: A factory function that returns a new instance of the
            configuration type.

            This function will be called when a new device of this type is added to the
            experiment.
            It should return a default instance of the configuration type.

        configuration_dumper: A function that converts a configuration instance to a
            JSON-serializable format.
            This function will be used to save the configuration.

        configuration_loader: A function that converts a JSON-serializable format to a
            configuration instance.
            This function will be used to load the configuration.
            It is passed the JSON-serializable format saved by the
            `configuration_dumper` function.

        editor_type: A function that creates an editor for the device configuration.

            When the user wants to edit the configuration of a device of this type,
            this function will be called to create an editor for the configuration.
            The function is passed as argument the configuration instance to edit.

            Once the user has finished editing the configuration, the method
            :meth:`DeviceConfiguration.get_configuration` of the editor is called and
            the result is saved.
    """

    label: str = attrs.field(converter=str)
    device_type: Callable[..., Device] = attrs.field()
    configuration_type: type[C] = attrs.field()
    configuration_factory: Callable[[], C] = attrs.field()
    configuration_dumper: Callable[[C], JSON] = attrs.field()
    configuration_loader: Callable[[JSON], C] = attrs.field()
    editor_type: Callable[[C], DeviceConfigurationEditor[C]] = attrs.field()
    compiler_type: type[DeviceCompiler] = attrs.field()
    controller_type: type[DeviceController] = attrs.field()
    proxy_type: type[DeviceProxy] = attrs.field()

    @configuration_type.validator  # type: ignore
    def _validate_configuration_type(self, _, value):
        if not issubclass(value, DeviceConfiguration):
            raise ValueError(f"Invalid configuration type: {value}.")

    @compiler_type.validator  # type: ignore
    def _validate_compiler_type(self, _, value):
        if not issubclass(value, DeviceCompiler):
            raise ValueError(f"Invalid compiler type: {value}.")

    @controller_type.validator  # type: ignore
    def _validate_controller_type(self, _, value):
        if not issubclass(value, DeviceController):
            raise ValueError(f"Invalid controller type: {value}.")

    @proxy_type.validator  # type: ignore
    def _validate_proxy_type(self, _, value):
        if not issubclass(value, DeviceProxy):
            raise ValueError(f"Invalid proxy type: {value}.")



#!/usr/bin/python
from PySide6 import QtWidgets, QtCore, QtGui

from .custom_widget_vectors import PropVector3, PropVector4
from .prop_widgets_abstract import BaseProperty


class PropColorPickerRGB(BaseProperty):
    """
    Color picker widget for a node property.
    """

    def __init__(self, parent=None):
        super(PropColorPickerRGB, self).__init__(parent)
        self._color = (0, 0, 0)
        self._button = QtWidgets.QPushButton()
        self._vector = PropVector3()
        self._vector.set_steps([1, 10, 100])
        self._vector.set_data_type(int)
        self._vector.set_value([0, 0, 0])
        self._vector.set_min(0)
        self._vector.set_max(255)
        self._update_color()

        self._button.clicked.connect(self._on_select_color)
        self._vector.value_changed.connect(self._on_vector_changed)

        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._button, 0, QtCore.Qt.AlignLeft)
        layout.addWidget(self._vector, 1, QtCore.Qt.AlignLeft)

    def _on_vector_changed(self, _, value):
        self._color = tuple(value)
        self._update_color()
        self.value_changed.emit(self.get_name(), value)

    def _on_select_color(self):
        current_color = QtGui.QColor(*self.get_value())
        color = QtWidgets.QColorDialog.getColor(current_color, self)
        if color.isValid():
            self.set_value(color.getRgb())

    def _update_vector(self):
        self._vector.set_value(self._color)

    def _update_color(self):
        c = [int(max(min(i, 255), 0)) for i in self._color]
        hex_color = "#{0:02x}{1:02x}{2:02x}".format(*c)
        self._button.setStyleSheet(
            """
            QPushButton {{background-color: rgba({0}, {1}, {2}, 255);}}
            QPushButton::hover {{background-color: rgba({0}, {1}, {2}, 200);}}
            """.format(
                *c
            )
        )
        self._button.setToolTip("rgb: {}\nhex: {}".format(self._color[:3], hex_color))

    def set_data_type(self, data_type):
        """
        Sets the input line edit fields to either display in float or int.

        Args:
            data_type(int or float): int or float data type object.
        """
        self._vector.set_data_type(data_type)

    def get_value(self):
        return self._color[:3]

    def set_value(self, value):
        if value != self.get_value():
            self._color = value
            self._update_color()
            self._update_vector()
            self.value_changed.emit(self.get_name(), value)


class PropColorPickerRGBA(PropColorPickerRGB):
    """
    Color4 (rgba) picker widget for a node property.
    """

    def __init__(self, parent=None):
        BaseProperty.__init__(self, parent)
        self._color = (0, 0, 0, 255)
        self._button = QtWidgets.QPushButton()
        self._vector = PropVector4()
        self._vector.set_steps([1, 10, 100])
        self._vector.set_data_type(int)
        self._vector.set_value([0, 0, 0, 255])
        self._vector.set_min(0)
        self._vector.set_max(255)
        self._update_color()

        self._button.clicked.connect(self._on_select_color)
        self._vector.value_changed.connect(self._on_vector_changed)

        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._button, 0, QtCore.Qt.AlignLeft)
        layout.addWidget(self._vector, 1, QtCore.Qt.AlignLeft)

    def _update_color(self):
        c = [int(max(min(i, 255), 0)) for i in self._color]
        hex_color = "#{0:02x}{1:02x}{2:02x}{3:03x}".format(*c)
        self._button.setStyleSheet(
            """
            QPushButton {{background-color: rgba({0}, {1}, {2}, {3});}}
            QPushButton::hover {{background-color: rgba({0}, {1}, {2}, {3});}}
            """.format(
                *c
            )
        )
        self._button.setToolTip("rgba: {}\nhex: {}".format(self._color, hex_color))

    def get_value(self):
        return self._color[:4]



import qtawesome
from PySide6.QtGui import QIcon, QPalette


def get_icon(name: str, color=None) -> QIcon:
    """Get an icon by name.

    Args:
        name: The name of the icon.
        color: The color of the icon. If None, the default color is used.
    """

    ids = {
        "camera": "mdi6.camera-outline",
        "editable-sequence": "mdi6.pencil-outline",
        "read-only-sequence": "mdi6.pencil-off-outline",
        "start": "mdi6.play",
        "stop": "mdi6.stop",
        "delete": "mdi6.delete",
        "duplicate": "mdi6.content-duplicate",
        "clear": "mdi6.database-remove",
        "plus": "mdi6.plus",
        "minus": "mdi6.minus",
        "copy": "mdi6.content-copy",
        "paste": "mdi6.content-paste",
        "simplify-timelanes": "mdi6.table-merge-cells",
        "add-time-lane": "mdi6.table-row-plus-after",
    }
    if color is None:
        color = QPalette().buttonText().color()
    icon = qtawesome.icon(ids[name], color=color)
    return icon



import abc
from collections.abc import Sequence, Iterable

import polars

from caqtus.session import Shot
from .shot_data import DataImporter


class CombinableLoader(DataImporter, abc.ABC):
    """A loader that can be combined with other loaders.

    Objects that inherit from this class can be combined with other loaders using the
    `+` and `*` operators.
    The `+` operator will concatenate the dataframes returned by the loaders.
    The `*` operator will perform a cross product of the dataframes returned by the
    loaders.
    """

    def __call__(self, shot: Shot) -> polars.DataFrame:
        return self.load(shot)

    @abc.abstractmethod
    def load(self, shot: Shot) -> polars.DataFrame: ...

    def __add__(self, other):
        if isinstance(other, CombinableLoader):
            return HorizontalConcatenateLoader(self, other)
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, CombinableLoader):
            return CrossProductLoader(self, other)
        else:
            return NotImplemented


class HorizontalConcatenateLoader(CombinableLoader):
    def __init__(self, *loaders: CombinableLoader):
        self.loaders = []
        for loader in loaders:
            if isinstance(loader, HorizontalConcatenateLoader):
                self.loaders.extend(loader.loaders)
            else:
                self.loaders.append(loader)

    @staticmethod
    def _concatenate(dataframes: Iterable[polars.DataFrame]) -> polars.DataFrame:
        return polars.concat(dataframes, how="horizontal")

    def load(self, shot: Shot) -> polars.DataFrame:
        return self._concatenate(loader(shot) for loader in self.loaders)


class CrossProductLoader(CombinableLoader):
    def __init__(self, first: CombinableLoader, second: CombinableLoader):
        self.first = first
        self.second = second

    @staticmethod
    def _join(first: polars.DataFrame, second: polars.DataFrame) -> polars.DataFrame:
        return first.join(second, how="cross")

    def load(self, shot: Shot) -> polars.DataFrame:
        return self._join(self.first(shot), self.second(shot))


# noinspection PyPep8Naming
class join(CombinableLoader):
    """Join multiple loaders on given columns."""

    def __init__(self, *loaders: CombinableLoader, on: Sequence[str]):
        if len(loaders) < 1:
            raise ValueError("At least one loader must be provided.")
        self.loaders = loaders
        self.on = on

    def _join(self, dataframes: Sequence[polars.DataFrame]) -> polars.DataFrame:
        dataframe = dataframes[0]
        for other in dataframes[1:]:
            dataframe = dataframe.join(other, on=self.on, how="inner")
        return dataframe

    def load(self, shot: Shot) -> polars.DataFrame:
        return self._join([loader(shot) for loader in self.loaders])



from typing import TypeVar, assert_never

from returns.result import Result, Success, Failure

T = TypeVar("T")
E = TypeVar("E", bound=Exception)


def unwrap(result: Result[T, E]) -> T:
    match result:
        case Success(value):
            return value
        case Failure(error):
            raise error
        case other:
            assert_never(other)


def is_success(result: Result[T, E]) -> bool:
    match result:
        case Success(_):
            return True
        case Failure(_):
            return False
        case other:
            assert_never(other)



from __future__ import annotations

import logging
from contextlib import AbstractContextManager
from typing import Protocol, TYPE_CHECKING

from caqtus.types.parameter import ParameterNamespace
from .device_configuration_collection import DeviceConfigurationCollection
from .path import PureSequencePath
from .path_hierarchy import PathHierarchy
from .sequence_collection import SequenceCollection

if TYPE_CHECKING:
    from .sequence import Sequence

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ExperimentSessionNotActiveError(RuntimeError):
    pass


class ExperimentSession(
    AbstractContextManager,
    Protocol,
):
    """Provides a connection to access the permanent storage of the experiment.

    An :py:class:`ExperimentSession` object allows to read and write configurations
    and data of the experiment.
    Every function and method that read or write data do so through an experiment
    session object.

    A session contains the following data:
    - A hierarchy of paths.
    - A collection of sequences.
    - A default collection of device configurations used to run a sequence.
    - A collection of global parameters.
    Global parameters are parameters that are not specific to a sequence, but are
    relevant for all the sequences.


    An experiment session object must be activated before it can be used.
    This is done by using the `with` statement on the session, inside which the session
    is active.
    If an error occurs inside the `with` block of the session, the data will be
    rolled back to the state it was in before the `with` block was entered in order to
    prevent leaving the storage in an inconsistent state.
    Data is only committed to the permanent storage when the `with` block is exited and
    will only be visible to other sessions after that point.
    For this reason, it is recommended to keep the `with` block as short as possible.

    A given session is not meant to be used concurrently.
    It can't be pickled and must not be passed to other processes.
    It is also not thread safe.
    It is not meant to be used by several coroutines at the same time, even if they
    belong to the same thread.

    It is possible to create multiple sessions connecting to the same storage using an
    :py:class:`caqtus.session.ExperimentSessionMaker`.
    """

    paths: PathHierarchy
    sequences: SequenceCollection
    default_device_configurations: DeviceConfigurationCollection

    def get_sequence(self, path: PureSequencePath | str) -> Sequence:
        """Get a sequence object from the session.

        Args:
            path: The path of the sequence to get.

        Returns:
            The sequence object.

        Raises:
            SequenceNotFoundError: If the sequence does not exist.
        """

        from .sequence import Sequence

        return Sequence(path, self)

    def get_global_parameters(self) -> ParameterNamespace:
        """Returns a copy of the global parameters of the session."""

        ...

    def set_global_parameters(self, parameters: ParameterNamespace) -> None:
        """Overwrite the global parameters of the session with the given parameters."""

        ...



import contextlib
from typing import TypeVar

import eliot

from caqtus.device.remote import DeviceProxy
from ._runtime import Camera

CameraType = TypeVar("CameraType", bound=Camera)


class CameraProxy(DeviceProxy[CameraType]):
    @eliot.log_call(include_args=["timeout"], include_result=False)
    async def update_parameters(self, timeout: float, *args, **kwargs) -> None:
        return await self.call_method("update_parameters", timeout, *args, **kwargs)

    @contextlib.asynccontextmanager
    async def acquire(self, exposures: list[float]):
        with eliot.start_action(action_type="acquire", exposures=exposures):
            async with self.call_method_proxy_result(
                "acquire", exposures
            ) as cm_proxy, self.async_context_manager(cm_proxy) as iterator_proxy:
                yield self.async_iterator(iterator_proxy)
                return



from typing import Literal

import attrs
import polars

from caqtus.session import Shot, Sequence
from caqtus.types.parameter import is_analog_value, is_quantity, Parameter
from caqtus.types.variable_name import DottedVariableName
from .combinable_importers import CombinableLoader
from .sequence_cache import cache_per_sequence


@attrs.define
class LoadShotParameters(CombinableLoader):
    """Loads the parameters of a shot.

    When it is evaluated on a shot, it returns a polars dataframe with a single row and
    with several columns named after each parameter requested.

    If some parameters are quantity with units, the dtype of the associated column will
    be a quantity dtype with two fields, magnitude and units.

    Attributes:
        which: the parameters to load from a shot.
            If it is "sequence", only the parameters defined at the sequence level are
            loaded.
            If "all", both sequence specific and global parameters are loaded.
    """

    which: Literal["sequence", "all"] = "sequence"

    def __attrs_post_init__(self):
        self._get_local_parameters = cache_per_sequence(get_local_parameters)

    @staticmethod
    def _parameters_to_dataframe(
        parameters: dict[DottedVariableName, Parameter]
    ) -> polars.DataFrame:
        series: list[polars.Series] = []

        for parameter_name, value in parameters.items():
            name = str(parameter_name)
            if is_analog_value(value) and is_quantity(value):
                magnitude = float(value.magnitude)
                units = format(value.units, "~")
                s = polars.Series(
                    name,
                    [
                        polars.Series("magnitude", [magnitude]),
                        polars.Series("units", [units], dtype=polars.Categorical),
                    ],
                    dtype=polars.Struct,
                )
            else:
                s = polars.Series(name, [value])
            series.append(s)
        series.sort(key=lambda s: s.name)
        dataframe = polars.DataFrame(series)
        return dataframe

    def load(self, shot: Shot) -> polars.DataFrame:
        parameters = shot.get_parameters()

        if self.which == "all":
            pass
        elif self.which == "sequence":
            local_parameters = self._get_local_parameters(shot.sequence)
            parameters = {name: parameters[name] for name in local_parameters}
        elif self.which == "globals":
            raise NotImplementedError
        else:
            raise NotImplementedError

        return self._parameters_to_dataframe(parameters)


def get_local_parameters(sequence: Sequence) -> set[DottedVariableName]:
    return sequence.get_local_parameters()
